"""Generates the supported SOP Classes and well-known SOP Instances."""

import inspect
import logging
import sys
from typing import Optional, Type, Any, cast, Dict

from pydicom.uid import UID

from pynetdicom.service_class import (
    BasicWorklistManagementServiceClass,
    ColorPaletteQueryRetrieveServiceClass,
    DefinedProcedureProtocolQueryRetrieveServiceClass,
    HangingProtocolQueryRetrieveServiceClass,
    ImplantTemplateQueryRetrieveServiceClass,
    NonPatientObjectStorageServiceClass,
    ProtocolApprovalQueryRetrieveServiceClass,
    QueryRetrieveServiceClass,
    RelevantPatientInformationQueryServiceClass,
    ServiceClass,
    StorageServiceClass,
    SubstanceAdministrationQueryServiceClass,
    VerificationServiceClass,
)
from pynetdicom.service_class_n import (
    ApplicationEventLoggingServiceClass,
    DisplaySystemManagementServiceClass,
    InstanceAvailabilityNotificationServiceClass,
    MediaCreationManagementServiceClass,
    PrintManagementServiceClass,
    ProcedureStepServiceClass,
    RTMachineVerificationServiceClass,
    StorageCommitmentServiceClass,
    UnifiedProcedureStepServiceClass,
)


LOGGER = logging.getLogger("pynetdicom.sop")


def uid_to_service_class(uid: str) -> Type[ServiceClass]:
    """Return the :class:`~pynetdicom.service_class.ServiceClass` object
    corresponding to `uid`.

    Parameters
    ----------
    uid : pydicom.uid.UID
        The SOP or Service Class UID to use to find the corresponding Service
        Class.

    Returns
    -------
    subclass of service_class.ServiceClass
        The Service Class corresponding to the SOP Class UID or the base class
        if support for the SOP Class isn't implemented.
    """
    if uid in _VERIFICATION_CLASSES.values():
        return VerificationServiceClass
    elif uid in _QR_CLASSES.values():
        return QueryRetrieveServiceClass
    elif uid in _STORAGE_CLASSES.values():
        return StorageServiceClass
    elif uid in _SERVICE_CLASSES:
        return _SERVICE_CLASSES[uid]
    elif uid in _APPLICATION_EVENT_CLASSES.values():
        return ApplicationEventLoggingServiceClass
    elif uid in _BASIC_WORKLIST_CLASSES.values():
        return BasicWorklistManagementServiceClass
    elif uid in _COLOR_PALETTE_CLASSES.values():
        return ColorPaletteQueryRetrieveServiceClass
    elif uid in _DEFINED_PROCEDURE_CLASSES.values():
        return DefinedProcedureProtocolQueryRetrieveServiceClass
    elif uid in _DISPLAY_SYSTEM_CLASSES.values():
        return DisplaySystemManagementServiceClass
    elif uid in _HANGING_PROTOCOL_CLASSES.values():
        return HangingProtocolQueryRetrieveServiceClass
    elif uid in _IMPLANT_TEMPLATE_CLASSES.values():
        return ImplantTemplateQueryRetrieveServiceClass
    elif uid in _INSTANCE_AVAILABILITY_CLASSES.values():
        return InstanceAvailabilityNotificationServiceClass
    elif uid in _MEDIA_CREATION_CLASSES.values():
        return MediaCreationManagementServiceClass
    elif uid in _MEDIA_STORAGE_CLASSES.values():
        return ServiceClass  # Not yet implemented
    elif uid in _NON_PATIENT_OBJECT_CLASSES.values():
        return NonPatientObjectStorageServiceClass
    elif uid in _PRINT_MANAGEMENT_CLASSES.values():
        return PrintManagementServiceClass
    elif uid in _PROCEDURE_STEP_CLASSES.values():
        return ProcedureStepServiceClass
    elif uid in _PROTOCOL_APPROVAL_CLASSES.values():
        return ProtocolApprovalQueryRetrieveServiceClass
    elif uid in _RELEVANT_PATIENT_QUERY_CLASSES.values():
        return RelevantPatientInformationQueryServiceClass
    elif uid in _RT_MACHINE_VERIFICATION_CLASSES.values():
        return RTMachineVerificationServiceClass
    elif uid in _STORAGE_COMMITMENT_CLASSES.values():
        return StorageCommitmentServiceClass
    elif uid in _SUBSTANCE_ADMINISTRATION_CLASSES.values():
        return SubstanceAdministrationQueryServiceClass
    elif uid in _UNIFIED_PROCEDURE_STEP_CLASSES.values():
        return UnifiedProcedureStepServiceClass

    # No SCP implemented
    return ServiceClass


class SOPClass(UID):
    """Extend :class:`~pydicom.uid.UID` to include the corresponding Service
    Class.

    """

    _service_class: Optional[Type[ServiceClass]] = None

    def __new__(cls: Type["SOPClass"], val: str) -> "SOPClass":
        if isinstance(val, SOPClass):
            return val

        return super(SOPClass, cls).__new__(cls, val)

    def __getattribute__(self, name: str) -> Any:
        return super(SOPClass, self).__getattribute__(name)

    @property
    def service_class(self) -> ServiceClass:
        """Return the corresponding Service Class implementation."""
        return cast(ServiceClass, self._service_class)


def _generate_sop_classes(sop_class_dict: Dict[str, str]) -> None:
    """Generate the SOP Classes."""
    for name in sop_class_dict:
        uid = sop_class_dict[name]
        sop_class: SOPClass = SOPClass(uid)
        sop_class._service_class = uid_to_service_class(uid)
        docstring = f"``{uid}``"
        # if uid in _x:
        #     docstring += "\n\n.. versionadded:: 1.4"

        sop_class.__doc__ = docstring
        globals()[name] = sop_class


# Table of service classes with assigned UIDs
_SERVICE_CLASSES = {
    "1.2.840.10008.4.2": StorageServiceClass,
    "1.2.840.10008.5.1.4.34.6": UnifiedProcedureStepServiceClass,
}

# Generate the various SOP classes
# pylint: disable=line-too-long
_APPLICATION_EVENT_CLASSES = {
    "ProceduralEventLogging": "1.2.840.10008.1.40",
    "SubstanceAdministrationLogging": "1.2.840.10008.1.42",
}
_BASIC_WORKLIST_CLASSES = {
    "ModalityWorklistInformationFind": "1.2.840.10008.5.1.4.31",
}
_COLOR_PALETTE_CLASSES = {
    "ColorPaletteInformationModelFind": "1.2.840.10008.5.1.4.39.2",
    "ColorPaletteInformationModelMove": "1.2.840.10008.5.1.4.39.3",
    "ColorPaletteInformationModelGet": "1.2.840.10008.5.1.4.39.4",
}
_DEFINED_PROCEDURE_CLASSES = {
    "DefinedProcedureProtocolInformationModelFind": "1.2.840.10008.5.1.4.20.1",
    "DefinedProcedureProtocolInformationModelMove": "1.2.840.10008.5.1.4.20.2",
    "DefinedProcedureProtocolInformationModelGet": "1.2.840.10008.5.1.4.20.3",
}
_DISPLAY_SYSTEM_CLASSES = {
    "DisplaySystem": "1.2.840.10008.5.1.1.40",
}
_HANGING_PROTOCOL_CLASSES = {
    "HangingProtocolInformationModelFind": "1.2.840.10008.5.1.4.38.2",
    "HangingProtocolInformationModelMove": "1.2.840.10008.5.1.4.38.3",
    "HangingProtocolInformationModelGet": "1.2.840.10008.5.1.4.38.4",
}
_IMPLANT_TEMPLATE_CLASSES = {
    "GenericImplantTemplateInformationModelFind": "1.2.840.10008.5.1.4.43.2",
    "GenericImplantTemplateInformationModelMove": "1.2.840.10008.5.1.4.43.3",
    "GenericImplantTemplateInformationModelGet": "1.2.840.10008.5.1.4.43.4",
    "ImplantAssemblyTemplateInformationModelFind": "1.2.840.10008.5.1.4.44.2",
    "ImplantAssemblyTemplateInformationModelMove": "1.2.840.10008.5.1.4.44.3",
    "ImplantAssemblyTemplateInformationModelGet": "1.2.840.10008.5.1.4.44.4",
    "ImplantTemplateGroupInformationModelFind": "1.2.840.10008.5.1.4.45.2",
    "ImplantTemplateGroupInformationModelMove": "1.2.840.10008.5.1.4.45.3",
    "ImplantTemplateGroupInformationModelGet": "1.2.840.10008.5.1.4.45.4",
}
_INSTANCE_AVAILABILITY_CLASSES = {
    "InstanceAvailabilityNotification": "1.2.840.10008.5.1.4.33",
}
_MEDIA_CREATION_CLASSES = {
    "MediaCreationManagement": "1.2.840.10008.5.1.1.33",
}
_MEDIA_STORAGE_CLASSES = {
    "MediaStorageDirectoryStorage": "1.2.840.10008.1.3.10",
}
_NON_PATIENT_OBJECT_CLASSES = {
    "HangingProtocolStorage": "1.2.840.10008.5.1.4.38.1",
    "ColorPaletteStorage": "1.2.840.10008.5.1.4.39.1",
    "GenericImplantTemplateStorage": "1.2.840.10008.5.1.4.43.1",
    "ImplantAssemblyTemplateStorage": "1.2.840.10008.5.1.4.44.1",
    "ImplantTemplateGroupStorage": "1.2.840.10008.5.1.4.45.1",
    "CTDefinedProcedureProtocolStorage": "1.2.840.10008.5.1.4.1.1.200.1",
    "ProtocolApprovalStorage": "1.2.840.10008.5.1.4.1.1.200.3",
    "XADefinedProcedureProtocolStorage": "1.2.840.10008.5.1.4.1.1.200.7",
}
_PRINT_MANAGEMENT_CLASSES = {
    "BasicFilmSession": "1.2.840.10008.5.1.1.1",
    "BasicFilmBox": "1.2.840.10008.5.1.1.2",
    "BasicGrayscaleImageBox": "1.2.840.10008.5.1.1.4",
    "BasicColorImageBox": "1.2.840.10008.5.1.1.4.1",
    "PrintJob": "1.2.840.10008.5.1.1.14",
    "BasicAnnotationBox": "1.2.840.10008.5.1.1.15",
    "Printer": "1.2.840.10008.5.1.1.16",
    "PrinterConfigurationRetrieval": "1.2.840.10008.5.1.1.16.376",
    "PresentationLUT": "1.2.840.10008.5.1.1.23",
    # Print Management Meta SOP Classes
    # Basic Film Session, Basic Film Box, Basic Grayscale, Printer
    "BasicGrayscalePrintManagementMeta": "1.2.840.10008.5.1.1.9",
    # Basic Film Session, Basic Film Box, Basic Color, Printer
    "BasicColorPrintManagementMeta": "1.2.840.10008.5.1.1.18",
}
_PROCEDURE_STEP_CLASSES = {
    "ModalityPerformedProcedureStep": "1.2.840.10008.3.1.2.3.3",
    "ModalityPerformedProcedureStepRetrieve": "1.2.840.10008.3.1.2.3.4",
    "ModalityPerformedProcedureStepNotification": "1.2.840.10008.3.1.2.3.5",
}
_PROTOCOL_APPROVAL_CLASSES = {
    "ProtocolApprovalInformationModelFind": "1.2.840.10008.5.1.4.1.1.200.4",
    "ProtocolApprovalInformationModelMove": "1.2.840.10008.5.1.4.1.1.200.5",
    "ProtocolApprovalInformationModelGet": "1.2.840.10008.5.1.4.1.1.200.6",
}
_QR_CLASSES = {
    "PatientRootQueryRetrieveInformationModelFind": "1.2.840.10008.5.1.4.1.2.1.1",
    "PatientRootQueryRetrieveInformationModelMove": "1.2.840.10008.5.1.4.1.2.1.2",
    "PatientRootQueryRetrieveInformationModelGet": "1.2.840.10008.5.1.4.1.2.1.3",
    "StudyRootQueryRetrieveInformationModelFind": "1.2.840.10008.5.1.4.1.2.2.1",
    "StudyRootQueryRetrieveInformationModelMove": "1.2.840.10008.5.1.4.1.2.2.2",
    "StudyRootQueryRetrieveInformationModelGet": "1.2.840.10008.5.1.4.1.2.2.3",
    "PatientStudyOnlyQueryRetrieveInformationModelFind": "1.2.840.10008.5.1.4.1.2.3.1",
    "PatientStudyOnlyQueryRetrieveInformationModelMove": "1.2.840.10008.5.1.4.1.2.3.2",
    "PatientStudyOnlyQueryRetrieveInformationModelGet": "1.2.840.10008.5.1.4.1.2.3.3",
    "CompositeInstanceRootRetrieveMove": "1.2.840.10008.5.1.4.1.2.4.2",
    "CompositeInstanceRootRetrieveGet": "1.2.840.10008.5.1.4.1.2.4.3",
    "CompositeInstanceRetrieveWithoutBulkDataGet": "1.2.840.10008.5.1.4.1.2.5.3",
}
_RELEVANT_PATIENT_QUERY_CLASSES = {
    "GeneralRelevantPatientInformationQuery": "1.2.840.10008.5.1.4.37.1",
    "BreastImagingRelevantPatientInformationQuery": "1.2.840.10008.5.1.4.37.2",
    "CardiacRelevantPatientInformationQuery": "1.2.840.10008.5.1.4.37.3",
}
_RT_MACHINE_VERIFICATION_CLASSES = {
    "RTConventionalMachineVerification": "1.2.840.10008.5.1.4.34.8",
    "RTIonMachineVerification": "1.2.840.10008.5.1.4.34.9",
}
_STORAGE_CLASSES = {
    "ComputedRadiographyImageStorage": "1.2.840.10008.5.1.4.1.1.1",  # A.2
    "DigitalXRayImageStorageForPresentation": "1.2.840.10008.5.1.4.1.1.1.1",  # A.26
    "DigitalXRayImageStorageForProcessing": "1.2.840.10008.5.1.4.1.1.1.1.1",  # A.26
    "DigitalMammographyXRayImageStorageForPresentation": "1.2.840.10008.5.1.4.1.1.1.2",  # A.27
    "DigitalMammographyXRayImageStorageForProcessing": "1.2.840.10008.5.1.4.1.1.1.2.1",  # A.27
    "DigitalIntraOralXRayImageStorageForPresentation": "1.2.840.10008.5.1.4.1.1.1.3",  # A.28
    "DigitalIntraOralXRayImageStorageForProcessing": "1.2.840.10008.5.1.4.1.1.1.3.1",  # A.28
    "CTImageStorage": "1.2.840.10008.5.1.4.1.1.2",  # A.3
    "EnhancedCTImageStorage": "1.2.840.10008.5.1.4.1.1.2.1",  # A.38
    "LegacyConvertedEnhancedCTImageStorage": "1.2.840.10008.5.1.4.1.1.2.2",  # A.70
    "UltrasoundMultiFrameImageStorage": "1.2.840.10008.5.1.4.1.1.3.1",  # A.7
    "MRImageStorage": "1.2.840.10008.5.1.4.1.1.4",  # A.4
    "EnhancedMRImageStorage": "1.2.840.10008.5.1.4.1.1.4.1",  # A.36.2
    "MRSpectroscopyStorage": "1.2.840.10008.5.1.4.1.1.4.2",  # A.36.3
    "EnhancedMRColorImageStorage": "1.2.840.10008.5.1.4.1.1.4.3",  # A.36.4
    "LegacyConvertedEnhancedMRImageStorage": "1.2.840.10008.5.1.4.1.1.4.4",  # A.71
    "UltrasoundImageStorage": "1.2.840.10008.5.1.4.1.1.6.1",  # A.6
    "EnhancedUSVolumeStorage": "1.2.840.10008.5.1.4.1.1.6.2",  # A.59
    "SecondaryCaptureImageStorage": "1.2.840.10008.5.1.4.1.1.7",  # A.8.1
    "MultiFrameSingleBitSecondaryCaptureImageStorage": "1.2.840.10008.5.1.4.1.1.7.1",  # A.8.2
    "MultiFrameGrayscaleByteSecondaryCaptureImageStorage": "1.2.840.10008.5.1.4.1.1.7.2",  # A.8.3
    "MultiFrameGrayscaleWordSecondaryCaptureImageStorage": "1.2.840.10008.5.1.4.1.1.7.3",  # A.8.4
    "MultiFrameTrueColorSecondaryCaptureImageStorage": "1.2.840.10008.5.1.4.1.1.7.4",  # A.8.5
    "TwelveLeadECGWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.1.1",  # A.34.3
    "GeneralECGWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.1.2",  # A.34.4
    "AmbulatoryECGWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.1.3",  # A.34.5
    "HemodynamicWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.2.1",  # A.34.6
    "CardiacElectrophysiologyWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.3.1",  # A.34.7
    "BasicVoiceAudioWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.4.1",  # A.34.2
    "GeneralAudioWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.4.2",  # A.34.10
    "ArterialPulseWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.5.1",  # A.34.8
    "RespiratoryWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.6.1",  # A.34.9
    "MultichannelRespiratoryWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.6.2",  # A.34.16
    "RoutineScalpElectroencephalogramWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.7.1",  # A.34.12
    "ElectromyogramWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.7.2",  # A.34.13
    "ElectrooculogramWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.7.3",  # A.34.14
    "SleepElectroencephalogramWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.7.4",  # A.34.15
    "BodyPositionWaveformStorage": "1.2.840.10008.5.1.4.1.1.9.8.1",  # A.34.17
    "GrayscaleSoftcopyPresentationStateStorage": "1.2.840.10008.5.1.4.1.1.11.1",  # A.33.1
    "ColorSoftcopyPresentationStateStorage": "1.2.840.10008.5.1.4.1.1.11.2",  # A.33.2
    "PseudoColorSoftcopyPresentationStageStorage": "1.2.840.10008.5.1.4.1.1.11.3",  # A.33.3
    "BlendingSoftcopyPresentationStateStorage": "1.2.840.10008.5.1.4.1.1.11.4",  # A.33.4
    "XAXRFGrayscaleSoftcopyPresentationStateStorage": "1.2.840.10008.5.1.4.1.1.11.5",  # A.33.6
    "GrayscalePlanarMPRVolumetricPresentationStateStorage": "1.2.840.10008.5.1.4.1.1.11.6",  # A.80.1
    "CompositingPlanarMPRVolumetricPresentationStateStorage": "1.2.840.10008.5.1.4.1.1.11.7",  # A.80.1
    "AdvancedBlendingPresentationStateStorage": "1.2.840.10008.5.1.4.1.1.11.8",  # A.33.7
    "VolumeRenderingVolumetricPresentationStateStorage": "1.2.840.10008.5.1.4.1.1.11.9",  # A.80.2
    "SegmentedVolumeRenderingVolumetricPresentationStateStorage": "1.2.840.10008.5.1.4.1.1.11.10",  # A.80.2
    "MultipleVolumeRenderingVolumetricPresentationStateStorage": "1.2.840.10008.5.1.4.1.1.11.11",  # A.80.2
    "XRayAngiographicImageStorage": "1.2.840.10008.5.1.4.1.1.12.1",  # A.14
    "EnhancedXAImageStorage": "1.2.840.10008.5.1.4.1.1.12.1.1",  # A.47
    "XRayRadiofluoroscopicImageStorage": "1.2.840.10008.5.1.4.1.1.12.2",  # A.16
    "EnhancedXRFImageStorage": "1.2.840.10008.5.1.4.1.1.12.2.1",  # A.48
    "XRay3DAngiographicImageStorage": "1.2.840.10008.5.1.4.1.1.13.1.1",  # A.53
    "XRay3DCraniofacialImageStorage": "1.2.840.10008.5.1.4.1.1.13.1.2",  # A.54
    "BreastTomosynthesisImageStorage": "1.2.840.10008.5.1.4.1.1.13.1.3",  # A.55
    "BreastProjectionXRayImageStorageForPresentation": "1.2.840.10008.5.1.4.1.1.13.1.4",  # A.74
    "BreastProjectionXRayImageStorageForProcessing": "1.2.840.10008.5.1.4.1.1.13.1.5",  # A.74
    "IntravascularOpticalCoherenceTomographyImageStorageForPresentation": "1.2.840.10008.5.1.4.1.1.14.1",  # A.66
    "IntravascularOpticalCoherenceTomographyImageStorageForProcessing": "1.2.840.10008.5.1.4.1.1.14.2",  # A.66
    "NuclearMedicineImageStorage": "1.2.840.10008.5.1.4.1.1.20",  # A.5
    "ParametricMapStorage": "1.2.840.10008.5.1.4.1.1.30",  # A.75
    "RawDataStorage": "1.2.840.10008.5.1.4.1.1.66",  # A.37
    "SpatialRegistrationStorage": "1.2.840.10008.5.1.4.1.1.66.1",  # A.39.1
    "SpatialFiducialsStorage": "1.2.840.10008.5.1.4.1.1.66.2",  # A.40
    "DeformableSpatialRegistrationStorage": "1.2.840.10008.5.1.4.1.1.66.3",  # A.39.2
    "SegmentationStorage": "1.2.840.10008.5.1.4.1.1.66.4",  # A.51
    "SurfaceSegmentationStorage": "1.2.840.10008.5.1.4.1.1.66.5",  # A.57
    "TractographyResultsStorage": "1.2.840.10008.5.1.4.1.1.66.6",  # A.78
    "RealWorldValueMappingStorage": "1.2.840.10008.5.1.4.1.1.67",  # A.46
    "SurfaceScanMeshStorage": "1.2.840.10008.5.1.4.1.1.68.1",  # A.68
    "SurfaceScanPointCloudStorage": "1.2.840.10008.5.1.4.1.1.68.2",  # A.69
    "VLEndoscopicImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.1",  # A.32.1
    "VideoEndoscopicImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.1.1",  # A.32.5
    "VLMicroscopicImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.2",  # A.32.2
    "VideoMicroscopicImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.2.1",  # A.32.6
    "VLSlideCoordinatesMicroscopicImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.3",  # A.32.3
    "VLPhotographicImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.4",  # A.32.4
    "VideoPhotographicImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.4.1",  # A.32.7
    "OphthalmicPhotography8BitImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.5.1",  # A.41
    "OphthalmicPhotography16BitImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.5.2",  # A.42
    "StereometricRelationshipStorage": "1.2.840.10008.5.1.4.1.1.77.1.5.3",  # A.43
    "OphthalmicTomographyImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.5.4",  # A.52
    "WideFieldOphthalmicPhotographyStereographicProjectionImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.5.5",  # A.76
    "WideFieldOphthalmicPhotography3DCoordinatesImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.5.6",  # A.77
    "OphthalmicOpticalCoherenceTomographyEnFaceImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.5.7",  # A.83
    "OphthlamicOpticalCoherenceTomographyBscanVolumeAnalysisStorage": "1.2.840.10008.5.1.4.1.1.77.1.5.8",  # A.84
    "VLWholeSlideMicroscopyImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.6",  # A.32.8
    "DermoscopicPhotographyImageStorage": "1.2.840.10008.5.1.4.1.1.77.1.7",  # A.32.11
    "LensometryMeasurementsStorage": "1.2.840.10008.5.1.4.1.1.78.1",  # A.60.1
    "AutorefractionMeasurementsStorage": "1.2.840.10008.5.1.4.1.1.78.2",  # A.60.2
    "KeratometryMeasurementsStorage": "1.2.840.10008.5.1.4.1.1.78.3",  # A.60.3
    "SubjectiveRefractionMeasurementsStorage": "1.2.840.10008.5.1.4.1.1.78.4",  # A.60.4
    "VisualAcuityMeasurementsStorage": "1.2.840.10008.5.1.4.1.1.78.5",  # A.60.5
    "SpectaclePrescriptionReportStorage": "1.2.840.10008.5.1.4.1.1.78.6",  # A.35.9
    "OphthalmicAxialMeasurementsStorage": "1.2.840.10008.5.1.4.1.1.78.7",  # A.60.6
    "IntraocularLensCalculationsStorage": "1.2.840.10008.5.1.4.1.1.78.8",  # A.60.7
    "MacularGridThicknessAndVolumeReportStorage": "1.2.840.10008.5.1.4.1.1.79.1",  # A.35.11
    "OphthalmicVisualFieldStaticPerimetryMeasurementsStorage": "1.2.840.10008.5.1.4.1.1.80.1",  # A.65
    "OphthalmicThicknessMapStorage": "1.2.840.10008.5.1.4.1.1.81.1",  # A.67
    "CornealTopographyMapStorage": "1.2.840.10008.5.1.4.1.1.82.1",  # A.73
    "BasicTextSRStorage": "1.2.840.10008.5.1.4.1.1.88.11",  # A.35.1
    "EnhancedSRStorage": "1.2.840.10008.5.1.4.1.1.88.22",  # A.35.2
    "ComprehensiveSRStorage": "1.2.840.10008.5.1.4.1.1.88.33",  # A.35.3
    "Comprehensive3DSRStorage": "1.2.840.10008.5.1.4.1.1.88.34",  # A.35.13
    "ExtensibleSRStorage": "1.2.840.10008.5.1.4.1.1.88.35",  # A.35.15
    "ProcedureLogStorage": "1.2.840.10008.5.1.4.1.1.88.40",  # A.35.7
    "MammographyCADSRStorage": "1.2.840.10008.5.1.4.1.1.88.50",  # A.35.5
    "KeyObjectSelectionDocumentStorage": "1.2.840.10008.5.1.4.1.1.88.59",  # A.35.4
    "ChestCADSRStorage": "1.2.840.10008.5.1.4.1.1.88.65",  # A.35.6
    "XRayRadiationDoseSRStorage": "1.2.840.10008.5.1.4.1.1.88.67",  # A.35.8
    "RadiopharmaceuticalRadiationDoseSRStorage": "1.2.840.10008.5.1.4.1.1.88.68",  # A.35.14
    "ColonCADSRStorage": "1.2.840.10008.5.1.4.1.1.88.69",  # A.35.10
    "ImplantationPlanSRStorage": "1.2.840.10008.5.1.4.1.1.88.70",  # A.35.12
    "AcquisitionContextSRStorage": "1.2.840.10008.5.1.4.1.1.88.71",  # A.35.16
    "SimplifiedAdultEchoSRStorage": "1.2.840.10008.5.1.4.1.1.88.72",  # A.35.17
    "PatientRadiationDoseSRStorage": "1.2.840.10008.5.1.4.1.1.88.73",  # A.35.18
    "PlannedImagingAgentAdministrationSRStorage": "1.2.840.10008.5.1.4.1.1.88.74",  # A.35.19
    "PerformedImagingAgentAdministrationSRStorage": "1.2.840.10008.5.1.4.1.1.88.75",  # A.35.20
    "EnhancedXRayRadiationDoseSRStorage": "1.2.840.10008.5.1.4.1.1.88.76",  # A.35.
    "ContentAssessmentResultsStorage": "1.2.840.10008.5.1.4.1.1.90.1",  # A.81
    "MicroscopyBulkSimpleAnnotationsStorage": "1.2.840.10008.5.1.4.1.1.91.1",
    "EncapsulatedPDFStorage": "1.2.840.10008.5.1.4.1.1.104.1",  # A.45.1
    "EncapsulatedCDAStorage": "1.2.840.10008.5.1.4.1.1.104.2",  # A.45.2
    "EncapsulatedSTLStorage": "1.2.840.10008.5.1.4.1.1.104.3",  # A.85.1
    "EncapsulatedOBJStorage": "1.2.840.10008.5.1.4.1.1.104.4",  # A.85.2
    "EncapsulatedMTLStorage": "1.2.840.10008.5.1.4.1.1.104.5",  # A.85.3
    "PositronEmissionTomographyImageStorage": "1.2.840.10008.5.1.4.1.1.128",  # A.21
    # 128 presentation context limit here
    "LegacyConvertedEnhancedPETImageStorage": "1.2.840.10008.5.1.4.1.1.128.1",  # A.72
    "EnhancedPETImageStorage": "1.2.840.10008.5.1.4.1.1.130",  # A.56
    "BasicStructuredDisplayStorage": "1.2.840.10008.5.1.4.1.1.131",  # A.33.5
    "CTPerformedProcedureProtocolStorage": "1.2.840.10008.5.1.4.1.1.200.2",  # A.82.1
    "XAPerformedProcedureProtocolStorage": "1.2.840.10008.5.1.4.1.1.200.8",
    "RTImageStorage": "1.2.840.10008.5.1.4.1.1.481.1",  # A.17
    "RTDoseStorage": "1.2.840.10008.5.1.4.1.1.481.2",  # A.18
    "RTStructureSetStorage": "1.2.840.10008.5.1.4.1.1.481.3",  # A.19
    "RTBeamsTreatmentRecordStorage": "1.2.840.10008.5.1.4.1.1.481.4",  # A.29
    "RTPlanStorage": "1.2.840.10008.5.1.4.1.1.481.5",  # A.20
    "RTBrachyTreatmentRecordStorage": "1.2.840.10008.5.1.4.1.1.481.6",  # A.20
    "RTTreatmentSummaryRecordStorage": "1.2.840.10008.5.1.4.1.1.481.7",  # A.31
    "RTIonPlanStorage": "1.2.840.10008.5.1.4.1.1.481.8",  # A.49
    "RTIonBeamsTreatmentRecordStorage": "1.2.840.10008.5.1.4.1.1.481.9",  # A.50
    "RTPhysicianIntentStorage": "1.2.840.10008.5.1.4.1.1.481.10",  # A.86.1.2
    "RTSegmentAnnotationStorage": "1.2.840.10008.5.1.4.1.1.481.11",  # A.86.1.3
    "RTRadiationSetStorage": "1.2.840.10008.5.1.4.1.1.481.12",  # A.86.1.4
    "CArmPhotonElectronRadiationStorage": "1.2.840.10008.5.1.4.1.1.481.13",  # A.86.1.5
    "TomotherapeuticRadiationStorage": "1.2.840.10008.5.1.4.1.1.481.14",  # A.86.1.6
    "RoboticArmRadiationStorage": "1.2.840.10008.5.1.4.1.1.481.15",  # A.86.1.7
    "RTRadiationRecordSetStorage": "1.2.840.10008.5.1.4.1.1.481.16",  # A.86.1.8
    "RTRadiationSalvageRecordStorage": "1.2.840.10008.5.1.4.1.1.481.17",  # A.86.1.9
    "TomotherapeuticRadiationRecordStorage": "1.2.840.10008.5.1.4.1.1.481.18",  # A.86.1.10
    "CArmPhotonElectronRadiationRecordStorage": "1.2.840.10008.5.1.4.1.1.481.19",  # A.86.1.11
    "RoboticArmRadiationRecordStorage": "1.2.840.10008.5.1.4.1.1.481.20",  # A.86.1.12
    "RTRadiationSetDeliveryInstructionStorage": "1.2.840.10008.5.1.4.1.1.481.21",
    "RTTreatmentPreparationStorage": "1.2.840.10008.5.1.4.1.1.481.22",
    "RTBeamsDeliveryInstructionStorage": "1.2.840.10008.5.1.4.34.7",  # A.64
    "RTBrachyApplicationSetupDeliveryInstructionsStorage": "1.2.840.10008.5.1.4.34.10",  # A.79
}
_STORAGE_COMMITMENT_CLASSES = {
    "StorageCommitmentPushModel": "1.2.840.10008.1.20.1",
}
_SUBSTANCE_ADMINISTRATION_CLASSES = {
    "ProductCharacteristicsQuery": "1.2.840.10008.5.1.4.41",
    "SubstanceApprovalQuery": "1.2.840.10008.5.1.4.42",
}
_UNIFIED_PROCEDURE_STEP_CLASSES = {
    "UnifiedProcedureStepPush": "1.2.840.10008.5.1.4.34.6.1",
    "UnifiedProcedureStepWatch": "1.2.840.10008.5.1.4.34.6.2",
    "UnifiedProcedureStepPull": "1.2.840.10008.5.1.4.34.6.3",
    "UnifiedProcedureStepEvent": "1.2.840.10008.5.1.4.34.6.4",
    "UnifiedProcedureStepQuery": "1.2.840.10008.5.1.4.34.6.5",
}
_VERIFICATION_CLASSES = {
    "Verification": "1.2.840.10008.1.1",
}

# pylint: enable=line-too-long
_generate_sop_classes(_APPLICATION_EVENT_CLASSES)
_generate_sop_classes(_BASIC_WORKLIST_CLASSES)
_generate_sop_classes(_COLOR_PALETTE_CLASSES)
_generate_sop_classes(_DEFINED_PROCEDURE_CLASSES)
_generate_sop_classes(_DISPLAY_SYSTEM_CLASSES)
_generate_sop_classes(_HANGING_PROTOCOL_CLASSES)
_generate_sop_classes(_IMPLANT_TEMPLATE_CLASSES)
_generate_sop_classes(_INSTANCE_AVAILABILITY_CLASSES)
_generate_sop_classes(_MEDIA_CREATION_CLASSES)
_generate_sop_classes(_MEDIA_STORAGE_CLASSES)
_generate_sop_classes(_NON_PATIENT_OBJECT_CLASSES)
_generate_sop_classes(_PRINT_MANAGEMENT_CLASSES)
_generate_sop_classes(_PROCEDURE_STEP_CLASSES)
_generate_sop_classes(_PROTOCOL_APPROVAL_CLASSES)
_generate_sop_classes(_QR_CLASSES)
_generate_sop_classes(_RELEVANT_PATIENT_QUERY_CLASSES)
_generate_sop_classes(_RT_MACHINE_VERIFICATION_CLASSES)
_generate_sop_classes(_STORAGE_CLASSES)
_generate_sop_classes(_STORAGE_COMMITMENT_CLASSES)
_generate_sop_classes(_SUBSTANCE_ADMINISTRATION_CLASSES)
_generate_sop_classes(_UNIFIED_PROCEDURE_STEP_CLASSES)
_generate_sop_classes(_VERIFICATION_CLASSES)


def uid_to_sop_class(uid: str) -> SOPClass:
    """Return the :class:`SOPClass` object corresponding to `uid`.

    Parameters
    ----------
    uid : pydicom.uid.UID
        Return the corresponding object for this UID.

    Returns
    -------
    sop_class.SOPClass subclass
        The SOP class corresponding to `uid`.

    Raises
    ------
    NotImplementedError
        If the SOP Class corresponding to the given UID has not been
        implemented.
    """
    # Get a list of all the class members of the current module
    members = inspect.getmembers(
        sys.modules[__name__], lambda mbr: isinstance(mbr, str)
    )

    for obj in members:
        if hasattr(obj[1], "service_class") and obj[1] == uid:
            return cast(SOPClass, obj[1])

    sop_class = SOPClass(uid)
    sop_class._service_class = ServiceClass

    return sop_class


# Well-known SOP Instance UIDs for the supported Service Classes
DisplaySystemInstance = UID("1.2.840.10008.5.1.1.40.1")
"""``1.2.840.10008.5.1.1.40.1``

.. versionadded:: 1.5
"""
PrinterConfigurationRetrievalInstance = UID("1.2.840.10008.5.1.1.17.376")
"""``1.2.840.10008.5.1.1.17.376``

.. versionadded:: 1.5
"""
PrinterInstance = UID("1.2.840.10008.5.1.1.17")
"""``1.2.840.10008.5.1.1.17``

.. versionadded:: 1.5
"""
ProceduralEventLoggingInstance = UID("1.2.840.10008.1.40.1")
"""``1.2.840.10008.1.40.1``

.. versionadded:: 1.5
"""
StorageCommitmentPushModelInstance = UID("1.2.840.10008.1.20.1.1")
"""``1.2.840.10008.1.20.1.1``

.. versionadded:: 1.5
"""
SubstanceAdministrationLoggingInstance = UID("1.2.840.10008.1.42.1")
"""``1.2.840.10008.1.42.1``

.. versionadded:: 1.5
"""
UPSFilteredGlobalSubscriptionInstance = UID("1.2.840.10008.5.1.4.34.5.1")
"""``1.2.840.10008.5.1.4.34.5.1``

.. versionadded:: 1.5
"""
UPSGlobalSubscriptionInstance = UID("1.2.840.10008.5.1.4.34.5")
"""``1.2.840.10008.5.1.4.34.5``

.. versionadded:: 1.5
"""



"""Unit tests for the DIMSE Message classes."""

from io import BytesIO
import logging

import pytest

from pydicom.dataset import Dataset
from pydicom.tag import Tag
from pydicom.uid import UID

from pynetdicom.dimse_messages import (
    C_STORE_RQ,
    C_STORE_RSP,
    DIMSEMessage,
    C_ECHO_RQ,
    C_ECHO_RSP,
    C_FIND_RQ,
    C_FIND_RSP,
    C_MOVE_RQ,
    C_MOVE_RSP,
    C_GET_RQ,
    C_GET_RSP,
    N_EVENT_REPORT_RQ,
    N_EVENT_REPORT_RSP,
    N_SET_RQ,
    N_SET_RSP,
    N_GET_RQ,
    N_GET_RSP,
    N_ACTION_RQ,
    N_ACTION_RSP,
    N_CREATE_RQ,
    N_CREATE_RSP,
    N_DELETE_RQ,
    N_DELETE_RSP,
    C_CANCEL_RQ,
    _COMMAND_SET_KEYWORDS,
)
from pynetdicom.dimse_primitives import (
    C_STORE,
    C_ECHO,
    C_GET,
    C_MOVE,
    C_FIND,
    N_EVENT_REPORT,
    N_GET,
    N_SET,
    N_ACTION,
    N_CREATE,
    N_DELETE,
    C_CANCEL,
)
from pynetdicom.dsutils import encode, decode
from pynetdicom.pdu_primitives import P_DATA
from pynetdicom.utils import pretty_bytes
from .encoded_dimse_msg import (
    c_echo_rq_cmd,
    c_echo_rsp_cmd,
    c_store_rq_cmd,
    c_store_ds,
    c_store_rsp_cmd,
    c_find_rq_cmd,
    c_find_rq_ds,
    c_find_rsp_cmd,
    c_find_rsp_ds,
    c_get_rq_cmd,
    c_get_rq_ds,
    c_get_rsp_cmd,
    c_get_rsp_ds,
    c_move_rq_cmd,
    c_move_rq_ds,
    c_move_rsp_cmd,
    c_move_rsp_ds,
    c_move_rsp_cmd_with_dup,
)
from .encoded_dimse_n_msg import (
    n_er_rq_cmd,
    n_er_rq_ds,
    n_er_rsp_cmd,
    n_er_rsp_ds,
    n_get_rq_cmd,
    n_get_rsp_cmd,
    n_get_rsp_ds,
    n_delete_rq_cmd,
    n_delete_rsp_cmd,
    n_action_rq_cmd,
    n_action_rq_ds,
    n_action_rsp_cmd,
    n_action_rsp_ds,
    n_create_rq_cmd,
    n_create_rq_ds,
    n_create_rsp_cmd,
    n_create_rsp_ds,
    n_set_rq_cmd,
    n_set_rq_ds,
    n_set_rsp_cmd,
    n_set_rsp_ds,
    n_create_rq_cmd_empty,
    n_set_rq_cmd_empty,
)

LOGGER = logging.getLogger("pynetdicom")
LOGGER.setLevel(logging.CRITICAL)


def print_nice_bytes(bytestream):
    """Nice output for bytestream."""
    str_list = pretty_bytes(
        bytestream, prefix="b'\\x", delimiter="\\x", items_per_line=10, suffix="' \\"
    )
    for string in str_list:
        print(string)


def test_command_set_keywords():
    """Check the keywords are valid and unique"""
    for kw_list in _COMMAND_SET_KEYWORDS.values():
        assert len(kw_list) == len(list(set(kw_list)))
        for kw in kw_list:
            Tag(kw)


class TestDIMSEMessage:
    """Test DIMSEMessage class"""

    def test_fragment_pdv(self):
        """Test that the PDV fragmenter is working correctly."""
        dimse_msg = C_STORE_RQ()
        frag = dimse_msg._generate_pdv_fragments

        result = []
        for fragment in frag(c_echo_rsp_cmd, 1000):
            result.append(fragment)
        assert len(result) == 1
        assert result[0] == c_echo_rsp_cmd
        assert isinstance(result[0], bytes)
        assert result[-1] != b""

        result = []
        for fragment in frag(c_echo_rsp_cmd, 10):
            result.append(fragment)
        assert len(result) == 20
        assert result[0] == c_echo_rsp_cmd[:4]
        assert isinstance(result[0], bytes)
        assert result[-1] != b""

        with pytest.raises(ValueError):
            next(frag(c_echo_rsp_cmd, 6))

    def test_fragment_pdv_zero(self):
        """Test that the PDV fragmenter works correctly for 0 max PDU."""
        dimse_msg = C_STORE_RQ()
        frag = dimse_msg._generate_pdv_fragments
        result = []
        for fragment in frag(c_echo_rsp_cmd, 0):
            result.append(fragment)

        assert len(result) == 1
        assert result[0] == c_echo_rsp_cmd

    def test_encode(self):
        """Test encoding of a DIMSE message."""
        primitive = C_STORE()
        primitive.MessageID = 7
        primitive.AffectedSOPClassUID = "1.1.1"
        primitive.AffectedSOPInstanceUID = "1.2.1"
        primitive.Priority = 0x02
        primitive.MoveOriginatorApplicationEntityTitle = "UNITTEST"
        primitive.MoveOriginatorMessageID = 3

        # Test encode without dataset
        dimse_msg = C_STORE_RQ()
        dimse_msg.primitive_to_message(primitive)
        p_data_list = []
        for pdata in dimse_msg.encode_msg(12, 16):
            p_data_list.append(pdata)
        assert p_data_list[0].presentation_data_value_list[0][1][0:1] == b"\x01"
        assert p_data_list[-1].presentation_data_value_list[0][1][0:1] == b"\x03"
        assert dimse_msg.context_id == 12

        # Test encode with dataset
        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube^HeNe"
        primitive.DataSet = BytesIO(encode(ds, True, True))

        dimse_msg = C_STORE_RQ()
        dimse_msg.primitive_to_message(primitive)
        p_data_list = []
        for pdata in dimse_msg.encode_msg(13, 10):
            p_data_list.append(pdata)

        assert p_data_list[0].presentation_data_value_list[0][1][0:1] == b"\x01"
        assert p_data_list[-1].presentation_data_value_list[0][1][0:1] == b"\x02"
        assert p_data_list[-2].presentation_data_value_list[0][1][0:1] == b"\x00"
        assert p_data_list[-10].presentation_data_value_list[0][1][0:1] == b"\x03"
        assert dimse_msg.context_id == 13

        p_data_list = []
        for pdata in dimse_msg.encode_msg(1, 31682):
            p_data_list.append(pdata)
        assert p_data_list[0].presentation_data_value_list[0][1] == c_store_rq_cmd
        assert p_data_list[1].presentation_data_value_list[0][1] == c_store_ds

    def test_encode_zero(self):
        """Test encoding with a 0 max pdu length."""
        primitive = C_STORE()
        primitive.MessageID = 7
        primitive.AffectedSOPClassUID = "1.1.1"
        primitive.AffectedSOPInstanceUID = "1.2.1"
        primitive.Priority = 0x02
        primitive.MoveOriginatorApplicationEntityTitle = "UNITTEST"
        primitive.MoveOriginatorMessageID = 3

        # Test encode without dataset
        dimse_msg = C_STORE_RQ()
        dimse_msg.primitive_to_message(primitive)
        p_data_list = []
        for pdata in dimse_msg.encode_msg(12, 0):
            p_data_list.append(pdata)
        assert p_data_list[0].presentation_data_value_list[0][1][0:1] == b"\x03"
        assert dimse_msg.context_id == 12

        # Test encode with dataset
        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube^HeNe"
        primitive.DataSet = BytesIO(encode(ds, True, True))

        dimse_msg = C_STORE_RQ()
        dimse_msg.primitive_to_message(primitive)
        p_data_list = []
        for pdata in dimse_msg.encode_msg(13, 0):
            p_data_list.append(pdata)

        assert p_data_list[0].presentation_data_value_list[0][1][0:1] == b"\x03"
        assert p_data_list[0].presentation_data_value_list[0][1] == c_store_rq_cmd
        assert p_data_list[1].presentation_data_value_list[0][1][0:1] == b"\x02"
        assert p_data_list[1].presentation_data_value_list[0][1] == c_store_ds
        assert dimse_msg.context_id == 13

    def test_decode(self):
        """Test decoding of a DIMSE message."""
        primitive = C_STORE()
        primitive.MessageID = 7
        primitive.AffectedSOPClassUID = "1.1.1"
        primitive.AffectedSOPInstanceUID = "1.2.1"
        primitive.Priority = 0x02
        primitive.MoveOriginatorApplicationEntityTitle = "UNITTEST"
        primitive.MoveOriginatorMessageID = 3
        ds = Dataset()
        ds.PatientID = "Test1101"
        ds.PatientName = "Tube^HeNe"
        primitive.DataSet = BytesIO(encode(ds, True, True))
        dimse_msg = C_STORE_RQ()
        dimse_msg.primitive_to_message(primitive)

        # CMD: (1x4, 3x1), DS: (0x1, 2x1)
        p_data = dimse_msg.encode_msg(12, 24)

        # Command set decoding
        dimse_msg.decode_msg(next(p_data))  # MCHB 1
        dimse_msg.decode_msg(next(p_data))  # MCHB 1
        dimse_msg.decode_msg(next(p_data))  # MCHB 1
        dimse_msg.decode_msg(next(p_data))  # MCHB 1
        dimse_msg.decode_msg(next(p_data))  # MCHB 3 - end of command set
        assert dimse_msg.__class__ == C_STORE_RQ

        # Test decoded command set
        cs = dimse_msg.command_set
        assert cs.CommandGroupLength == 94
        assert cs.AffectedSOPClassUID == UID("1.1.1")
        assert cs.AffectedSOPInstanceUID == UID("1.2.1")
        assert cs.Priority == 2
        assert cs.CommandDataSetType == 1
        assert cs.CommandField == 1
        assert cs.MoveOriginatorApplicationEntityTitle == "UNITTEST"
        assert cs.MoveOriginatorMessageID == 3

        # Test decoded dataset
        dimse_msg.decode_msg(next(p_data))  # MCHB 1 # MCHB 0
        dimse_msg.decode_msg(next(p_data))  # MCHB 1 # MCHB 2
        assert dimse_msg.data_set.getvalue() == c_store_ds[1:]

        # Test returns false
        msg = C_STORE_RSP()
        assert not msg.decode_msg(c_store_rsp_cmd)

    def test_primitive_to_message(self):
        """Test converting a DIMSE primitive to a DIMSE message."""
        primitive = C_STORE()
        primitive.MessageID = 7
        primitive.AffectedSOPClassUID = "1.1.1"
        primitive.AffectedSOPInstanceUID = "1.2.1"
        primitive.Priority = 0x02
        msg = C_STORE_RQ()
        msg.primitive_to_message(primitive)
        # Test unused command set elements are deleted
        assert not "MoveOriginatorApplicationEntityTitle" in msg.command_set

        # Test raise error for unknown DIMSE message type
        msg.__class__.__name__ = "TestClass"
        with pytest.raises(ValueError):
            msg.primitive_to_message(primitive)

        # Reset name to avoid errors in other tests
        msg.__class__.__name__ = "C_STORE_RQ"

    # DIMSE-C
    def test_message_to_primitive_c_store(self):
        """Test converting C_STORE_RQ and _RSP to C_STORE primitive."""
        msg = C_STORE_RQ()
        for data in [c_store_rq_cmd, c_store_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_STORE)
        assert isinstance(primitive.DataSet, BytesIO)
        assert primitive.AffectedSOPClassUID == UID("1.1.1")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.1")
        assert primitive.Priority == 2
        assert primitive.MoveOriginatorApplicationEntityTitle == "UNITTEST"
        assert primitive.MoveOriginatorMessageID == 3

        ds = decode(primitive.DataSet, True, True)
        assert ds.PatientName == "Tube^HeNe"
        assert ds.PatientID == "Test1101"

        msg = C_STORE_RSP()
        p_data = P_DATA()
        p_data.presentation_data_value_list.append([0, c_store_rsp_cmd])
        msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_STORE)
        assert primitive.DataSet is None
        for elem in msg.command_set:
            if hasattr(primitive, elem.keyword):
                item = getattr(primitive, elem.keyword)
                assert item == elem.value

    def test_message_to_primitive_c_echo(self):
        """Test converting C_ECHO_RQ to C_ECHO primitive."""
        msg = C_ECHO_RQ()
        p_data = P_DATA()
        p_data.presentation_data_value_list.append([0, c_echo_rq_cmd])
        msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_ECHO)
        assert primitive.AffectedSOPClassUID == UID("1.2.840.10008.1.1")
        assert primitive.MessageID == 7

        msg = C_ECHO_RSP()
        p_data = P_DATA()
        p_data.presentation_data_value_list.append([0, c_echo_rsp_cmd])
        msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_ECHO)
        assert primitive.AffectedSOPClassUID == UID("1.2.840.10008.1.1")
        assert primitive.MessageIDBeingRespondedTo == 8
        assert primitive.Status == 0

    def test_message_to_primitive_c_get(self):
        """Test converting C_GET_RQ to C_GET primitive."""
        msg = C_GET_RQ()
        for data in [c_get_rq_cmd, c_get_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_GET)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID("1.2.840.10008.5.1.4.1.1.2")
        assert primitive.Priority == 2
        assert primitive.MessageID == 7

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == "PATIENT"
        assert ds.PatientID == "*"

        msg = C_GET_RSP()
        for data in [c_get_rsp_cmd, c_get_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_GET)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID("1.2.840.10008.5.1.4.1.1.2")
        assert primitive.Status == 65280
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.NumberOfRemainingSuboperations == 3
        assert primitive.NumberOfCompletedSuboperations == 1
        assert primitive.NumberOfFailedSuboperations == 2
        assert primitive.NumberOfWarningSuboperations == 4

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == "PATIENT"
        assert ds.PatientID == "*"

    def test_message_to_primitive_c_move(self):
        """Test converting C_MOVE_RQ to C_MOVE primitive."""
        msg = C_MOVE_RQ()
        for data in [c_move_rq_cmd, c_move_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_MOVE)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID("1.2.840.10008.5.1.4.1.1.2")
        assert primitive.Priority == 2
        assert primitive.MoveDestination == "MOVE_SCP"
        assert primitive.MessageID == 7

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == "PATIENT"
        assert ds.PatientID == "*"

        msg = C_MOVE_RSP()
        for data in [c_move_rsp_cmd, c_move_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_MOVE)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID("1.2.840.10008.5.1.4.1.1.2")
        assert primitive.Status == 65280
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.NumberOfRemainingSuboperations == 3
        assert primitive.NumberOfCompletedSuboperations == 1
        assert primitive.NumberOfFailedSuboperations == 2
        assert primitive.NumberOfWarningSuboperations == 4

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == "PATIENT"
        assert ds.PatientID == "*"

    def test_message_to_primitive_c_move_with_duplicate_command_fields(self):
        msg = C_MOVE_RSP()
        for data in [c_move_rsp_cmd_with_dup, c_move_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_MOVE)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID("1.2.840.10008.5.1.4.1.1.2")
        assert primitive.Status == 65280
        assert primitive.MessageIDBeingRespondedTo == 5

        # Fields with VM > 1 that can only appear once
        assert primitive.NumberOfRemainingSuboperations == 3
        assert primitive.NumberOfCompletedSuboperations == 1
        assert primitive.NumberOfFailedSuboperations == 2
        assert primitive.NumberOfWarningSuboperations == 4

        # Fields with VM > 1 that are allowed to appear multiple times
        assert primitive.OffendingElement == [(0x0, 0x1), (0x0, 0x2)]

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == "PATIENT"
        assert ds.PatientID == "*"

    def test_message_to_primitive_c_find(self):
        """Test converting C_FIND_RQ to C_FIND primitive."""
        msg = C_FIND_RQ()
        for data in [c_find_rq_cmd, c_find_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_FIND)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID("1.2.840.10008.5.1.4.1.1.2")
        assert primitive.Priority == 2
        assert primitive.MessageID == 7

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == "PATIENT"
        assert ds.PatientID == "*"

        msg = C_GET_RSP()
        for data in [c_find_rsp_cmd, c_find_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_FIND)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID("1.2.840.10008.5.1.4.1.1.2")
        assert primitive.Status == 65280

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == "PATIENT"
        assert ds.RetrieveAETitle == "FINDSCP"

        assert ds.PatientName == "ANON^A^B^C^D"

    def test_message_to_primitive_c_cancel(self):
        """Test converting C_CANCEL_rq to C_CANCEL primitive."""
        msg = C_CANCEL_RQ()
        cs = Dataset()
        cs.MessageIDBeingRespondedTo = 1
        msg.command_set = cs
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_CANCEL)
        assert primitive.MessageIDBeingRespondedTo == 1

    # DIMSE-N
    def test_message_to_primitive_n_event_report(self):
        """Test converting N_EVENT_REPORT_RQ and _RSP to primitive."""
        # N-EVENT-REPORT-RQ
        msg = N_EVENT_REPORT_RQ()
        for data in [n_er_rq_cmd, n_er_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_EVENT_REPORT)
        assert primitive.AffectedSOPClassUID == UID("1.2.840.10008.5.1.4.1.1.2")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.392.200036.9116.2.6.1.48")
        assert primitive.MessageID == 7
        assert primitive.EventTypeID == 2

        ds = decode(primitive.EventInformation, True, True)
        assert ds.PatientName == "Tube HeNe"
        assert ds.PatientID == "Test1101"

        # N-EVENT-REPORT-RSP
        msg = N_EVENT_REPORT_RSP()
        for data in [n_er_rsp_cmd, n_er_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_EVENT_REPORT)
        assert primitive.AffectedSOPClassUID == UID("1.2.4.10")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.4.5.7.8")
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.EventTypeID == 2
        assert primitive.Status == 0x0000

        ds = decode(primitive.EventReply, True, True)
        assert ds.PatientName == "Tube HeNe"
        assert ds.PatientID == "Test1101"

    def test_message_to_primitive_n_get(self):
        """Test converting N_GET_RQ and _RSP to primitive."""
        # N-GET-RQ
        msg = N_GET_RQ()
        for data in [n_get_rq_cmd]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_GET)
        assert primitive.RequestedSOPClassUID == UID("1.2.840.10008.5.1.4.1.1.2")
        assert primitive.RequestedSOPInstanceUID == UID("1.2.392.200036.9116.2.6.1.48")
        assert primitive.MessageID == 7
        primitive.AttributeIdentifierList = [
            (0x7FE0, 0x0010),
            (0x0000, 0x0000),
            (0xFFFF, 0xFFFF),
        ]

        msg = N_GET_RSP()
        for data in [n_get_rsp_cmd, n_get_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_GET)
        assert primitive.AttributeIdentifierList is None
        assert primitive.AffectedSOPClassUID == UID("1.2.4.10")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.4.5.7.8")
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.Status == 0x0000

        ds = decode(primitive.AttributeList, True, True)
        assert ds.PatientName == "Tube HeNe"
        assert ds.PatientID == "Test1101"

    def test_message_to_primitive_n_set(self):
        """Test converting N_SET_RQ and _RSP to primitive."""
        # N-SET-RQ
        msg = N_SET_RQ()
        for data in [n_set_rq_cmd, n_set_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_SET)
        assert primitive.RequestedSOPClassUID == UID("1.2.840.10008.5.1.4.1.1.2")
        assert primitive.RequestedSOPInstanceUID == UID("1.2.392.200036.9116.2.6.1.48")
        assert primitive.MessageID == 7

        ds = decode(primitive.ModificationList, True, True)
        assert ds.PatientName == "Tube HeNe"
        assert ds.PatientID == "Test1101"

        # N-SET-RSP
        msg = N_SET_RSP()
        for data in [n_set_rsp_cmd, n_set_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_SET)
        assert isinstance(primitive.AttributeList, BytesIO)
        assert primitive.AffectedSOPClassUID == UID("1.2.4.10")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.4.5.7.8")
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.Status == 0x0000

        ds = decode(primitive.AttributeList, True, True)
        assert ds.PatientName == "Tube HeNe"
        assert ds.PatientID == "Test1101"

    def test_message_to_primitive_n_action(self):
        """Test converting N_ACTION_RQ and _RSP to primitive."""
        # N-ACTION-RQ
        msg = N_ACTION_RQ()
        for data in [n_action_rq_cmd, n_action_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_ACTION)
        assert primitive.RequestedSOPClassUID == UID("1.2.840.10008.5.1.4.1.1.2")
        assert primitive.RequestedSOPInstanceUID == UID("1.2.392.200036.9116.2.6.1.48")
        assert primitive.MessageID == 7
        assert primitive.ActionTypeID == 1

        ds = decode(primitive.ActionInformation, True, True)
        assert ds.PatientName == "Tube HeNe"
        assert ds.PatientID == "Test1101"

        # N-ACTION-RSP
        msg = N_ACTION_RSP()
        for data in [n_action_rsp_cmd, n_action_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_ACTION)
        assert primitive.AffectedSOPClassUID == UID("1.2.4.10")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.4.5.7.8")
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.ActionTypeID == 1
        assert primitive.Status == 0x0000

        ds = decode(primitive.ActionReply, True, True)
        assert ds.PatientName == "Tube HeNe"
        assert ds.PatientID == "Test1101"

    def test_message_to_primitive_n_create(self):
        """Test converting N_CREATE_RQ and _RSP to primitive."""
        # N-CREATE-RQ
        msg = N_CREATE_RQ()
        for data in [n_create_rq_cmd, n_create_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_CREATE)
        assert primitive.AffectedSOPClassUID == UID("1.2.840.10008.5.1.4.1.1.2")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.392.200036.9116.2.6.1.48")
        assert primitive.MessageID == 7

        ds = decode(primitive.AttributeList, True, True)
        assert ds.PatientName == "Tube HeNe"
        assert ds.PatientID == "Test1101"

        # N-ACTION-RSP
        msg = N_CREATE_RSP()
        for data in [n_create_rsp_cmd, n_create_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_CREATE)
        assert primitive.AffectedSOPClassUID == UID("1.2.4.10")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.4.5.7.8")
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.Status == 0x0000

        ds = decode(primitive.AttributeList, True, True)
        assert ds.PatientName == "Tube HeNe"
        assert ds.PatientID == "Test1101"

    def test_message_to_primitive_n_delete(self):
        """Test converting N_DELETE_RQ and _RSP to primitive."""
        # N-DELETE-RQ
        msg = N_DELETE_RQ()
        for data in [n_delete_rq_cmd]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_DELETE)
        assert primitive.RequestedSOPClassUID == UID("1.2.3")
        assert primitive.RequestedSOPInstanceUID == UID("1.2.30")
        assert primitive.MessageID == 7

        # N-DELETE-RSP
        msg = N_DELETE_RSP()
        for data in [n_delete_rsp_cmd]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_DELETE)
        assert primitive.AffectedSOPClassUID == UID("1.2.4.10")
        assert primitive.AffectedSOPInstanceUID == UID("1.2.4.5.7.8")
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.Status == 0xC201

    def test_empty_uid_n_create(self):
        """Test converting N_CREATE_RQ with empty UID value."""
        msg = N_CREATE_RQ()
        for data in [n_create_rq_cmd_empty]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)

        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_CREATE)
        assert primitive.AffectedSOPClassUID == UID("1.2.3.4")
        assert primitive.MessageID == 7
        assert primitive.AffectedSOPInstanceUID is None

    def test_empty_uid_n_set(self):
        """Test converting N_CREATE_RQ with empty UID value."""
        msg = N_SET_RQ()
        for data in [n_set_rq_cmd_empty]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)

        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_SET)
        assert primitive.RequestedSOPClassUID is None
        assert primitive.MessageID == 7
        assert primitive.RequestedSOPInstanceUID is None


class TestThreadSafety:
    """Tests for the thread safety of DIMSEMessage classes."""

    def setup(self):
        ds = Dataset()
        ds.SOPClassUID = "1.2.840.10008.5.1.4.1.1"
        ds.SOPInstanceUID = "1.2.3.4"
        ds.PatientName = "Test^Testing"
        self.ds = ds

    def test_msg_init(self):
        """Test the DIMSE Message dynamic construction works OK."""
        msg = C_STORE_RQ()
        cs = msg.command_set
        ds = msg.data_set
        assert cs.CommandGroupLength is None
        assert cs.AffectedSOPClassUID in ["", None]  # pydicom upstream change
        assert cs.CommandField is None
        assert cs.MessageID is None
        assert cs.Priority is None
        assert cs.CommandDataSetType is None
        assert cs.AffectedSOPInstanceUID in ["", None]
        assert cs.MoveOriginatorApplicationEntityTitle is None
        assert cs.MoveOriginatorMessageID is None
        assert ds.getvalue() == b""

    def test_cstore_rq(self):
        """Test creation of a DIMSE Message from a primitive works OK."""
        # Testing that primitive_to_message doesn't affect the original class
        #   definition
        req = C_STORE()
        req.MessageID = 12
        req.AffectedSOPClassUID = self.ds.SOPClassUID
        req.AffectedSOPInstanceUID = self.ds.SOPInstanceUID
        req.Priority = 0x02
        req.MoveOriginatorApplicationEntityTitle = None
        req.MoveOriginatorMessageID = None
        req.DataSet = BytesIO(encode(self.ds, True, True))

        msg = C_STORE_RQ()
        msg.primitive_to_message(req)
        cs = msg.command_set
        assert cs.CommandGroupLength == 88
        assert cs.AffectedSOPClassUID == "1.2.840.10008.5.1.4.1.1"
        assert cs.CommandField == 1
        assert cs.MessageID == 12
        assert cs.Priority == 2
        assert cs.CommandDataSetType == 1
        assert cs.AffectedSOPInstanceUID == "1.2.3.4"
        assert "MoveOriginatorApplicationEntityTitle" not in cs
        assert "MoveOriginatorMessageID" not in cs
        assert msg.data_set.getvalue() != b""

        # Test new class instance is OK
        msg = C_STORE_RQ()
        cs = msg.command_set
        ds = msg.data_set
        assert cs.CommandGroupLength is None
        assert cs.AffectedSOPClassUID in ["", None]  # pydicom upstream change
        assert cs.CommandField is None
        assert cs.MessageID is None
        assert cs.Priority is None
        assert cs.CommandDataSetType is None
        assert cs.AffectedSOPInstanceUID in ["", None]
        assert cs.MoveOriginatorApplicationEntityTitle is None
        assert cs.MoveOriginatorMessageID is None
        assert ds.getvalue() == b""

    def test_message_builder_regression(self):
        """Regression test for DIMSEMessage class builder."""
        with pytest.raises(AttributeError, match=r"no attribute 'command_set'"):
            assert C_STORE_RQ.command_set.MessageID
        with pytest.raises(AttributeError, match=r"no attribute 'data_set'"):
            assert C_STORE_RQ.data_set.get_value() == b""



"""Unit tests for echoscu.py"""

import logging
import os
import subprocess
import sys
import time

import pytest

from pydicom.uid import (
    ExplicitVRLittleEndian,
    ImplicitVRLittleEndian,
    DeflatedExplicitVRLittleEndian,
    ExplicitVRBigEndian,
)

from pynetdicom import AE, evt, debug_logger, DEFAULT_TRANSFER_SYNTAXES
from pynetdicom.sop_class import Verification, CTImageStorage


# debug_logger()


APP_DIR = os.path.join(os.path.dirname(__file__), "../")
APP_FILE = os.path.join(APP_DIR, "echoscu", "echoscu.py")


def start_echoscu(args):
    """Start the echoscu.py app and return the process."""
    pargs = [sys.executable, APP_FILE, "localhost", "11112"] + [*args]
    return subprocess.Popen(pargs)


def start_echoscu_cli(args):
    """Start the echoscu app using CLI and return the process."""
    pargs = [sys.executable, "-m", "pynetdicom", "echoscu", "localhost", "11112"] + [
        *args
    ]
    return subprocess.Popen(pargs)


class EchoSCUBase:
    """Tests for echoscu.py"""

    def setup(self):
        """Run prior to each test"""
        self.ae = None
        self.func = None

    def teardown(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_default(self):
        """Test default settings."""
        events = []

        def handle_echo(event):
            events.append(event)
            return 0x0000

        def handle_release(event):
            events.append(event)

        handlers = [(evt.EVT_C_ECHO, handle_echo), (evt.EVT_RELEASED, handle_release)]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func([])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_ECHO
        assert events[1].event == evt.EVT_RELEASED
        requestor = events[1].assoc.requestor
        assert "ECHOSCU" == requestor.ae_title
        assert 16382 == requestor.maximum_length
        assert "ANY-SCP" == requestor.primitive.called_ae_title
        assert [] == requestor.extended_negotiation
        assert (1, 1) == requestor.asynchronous_operations
        assert {} == requestor.sop_class_common_extended
        assert {} == requestor.sop_class_extended
        assert requestor.user_identity == None
        cxs = requestor.primitive.presentation_context_definition_list
        assert len(cxs) == 1
        assert cxs[0].abstract_syntax == Verification
        assert cxs[0].transfer_syntax == [
            ExplicitVRLittleEndian,
            ImplicitVRLittleEndian,
            DeflatedExplicitVRLittleEndian,
            ExplicitVRBigEndian,
        ]

    def test_no_peer(self, capfd):
        """Test trying to connect to non-existent host."""
        p = self.func([])
        p.wait()
        assert p.returncode == 1

        out, err = capfd.readouterr()
        assert "Association request failed: unable to connect to remote" in err
        assert "TCP Initialisation Error" in err

    def test_flag_version(self, capfd):
        """Test --version flag."""
        p = self.func(["--version"])
        p.wait()
        assert p.returncode == 0

        out, err = capfd.readouterr()
        assert "echoscu.py v" in out

    def test_flag_quiet(self, capfd):
        """Test --quiet flag."""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(CTImageStorage)
        scp = ae.start_server(("localhost", 11112), block=False)

        p = self.func(["-q"])
        p.wait()
        assert p.returncode == 1

        out, err = capfd.readouterr()
        assert out == err == ""

        scp.shutdown()

    def test_flag_verbose(self, capfd):
        """Test --verbose flag."""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)

        p = self.func(["-v"])
        p.wait()
        assert p.returncode == 0

        out, err = capfd.readouterr()
        assert "Requesting Association" in err
        assert "Association Accepted" in err
        assert "Sending Echo Request" in err
        assert "Received Echo Response" in err
        assert "Releasing Association" in err
        assert "Accept Parameters" not in err

        scp.shutdown()

    def test_flag_debug(self, capfd):
        """Test --debug flag."""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)

        p = self.func(["-d"])
        p.wait()
        assert p.returncode == 0

        out, err = capfd.readouterr()
        assert "Releasing Association" in err
        assert "Accept Parameters" in err

        scp.shutdown()

    def test_flag_log_collision(self):
        """Test error with -q -v and -d flag."""
        p = self.func(["-q", "-v"])
        p.wait()
        assert p.returncode != 0

    @pytest.mark.skip("No way to test comprehensively")
    def test_flag_log_level(self):
        """Test --log-level flag."""
        pass

    def test_flag_aet(self):
        """Test --calling-aet flag."""
        events = []

        def handle_echo(event):
            events.append(event)
            return 0x0000

        def handle_release(event):
            events.append(event)

        handlers = [(evt.EVT_C_ECHO, handle_echo), (evt.EVT_RELEASED, handle_release)]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-aet", "MYSCU"])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_ECHO
        assert events[1].event == evt.EVT_RELEASED
        requestor = events[1].assoc.requestor
        assert "MYSCU" == requestor.ae_title

    def test_flag_aec(self):
        """Test --called-aet flag."""
        events = []

        def handle_echo(event):
            events.append(event)
            return 0x0000

        def handle_release(event):
            events.append(event)

        handlers = [(evt.EVT_C_ECHO, handle_echo), (evt.EVT_RELEASED, handle_release)]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-aec", "YOURSCP"])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_ECHO
        assert events[1].event == evt.EVT_RELEASED
        requestor = events[1].assoc.requestor
        assert "YOURSCP" == requestor.primitive.called_ae_title

    def test_flag_ta(self, capfd):
        """Test --acse-timeout flag."""
        events = []

        def handle_requested(event):
            events.append(event)
            time.sleep(0.1)

        def handle_echo(event):
            events.append(event)
            return 0x0000

        def handle_abort(event):
            events.append(event)

        handlers = [
            (evt.EVT_C_ECHO, handle_echo),
            (evt.EVT_ABORTED, handle_abort),
            (evt.EVT_REQUESTED, handle_requested),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-ta", "0.05", "-d"])
        p.wait()
        assert p.returncode == 1

        time.sleep(0.1)

        scp.shutdown()

        out, err = capfd.readouterr()
        assert "ACSE timeout reached while waiting for response" in err
        assert events[0].event == evt.EVT_REQUESTED
        assert events[1].event == evt.EVT_ABORTED

    def test_flag_td(self, capfd):
        """Test --dimse-timeout flag."""
        events = []

        def handle_echo(event):
            events.append(event)
            time.sleep(0.1)
            return 0x0000

        def handle_abort(event):
            events.append(event)

        handlers = [
            (evt.EVT_C_ECHO, handle_echo),
            (evt.EVT_ABORTED, handle_abort),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-td", "0.05", "-d"])
        p.wait()
        assert p.returncode == 0

        time.sleep(0.1)

        scp.shutdown()

        out, err = capfd.readouterr()
        assert "DIMSE timeout reached while waiting for message" in err
        assert events[0].event == evt.EVT_C_ECHO
        assert events[1].event == evt.EVT_ABORTED

    @pytest.mark.skip("Don't think this can be tested")
    def test_flag_tn(self, capfd):
        """Test --network-timeout flag."""
        pass

    def test_flag_max_pdu(self):
        """Test --max-pdu flag."""
        events = []

        def handle_echo(event):
            events.append(event)
            return 0x0000

        def handle_release(event):
            events.append(event)

        handlers = [(evt.EVT_C_ECHO, handle_echo), (evt.EVT_RELEASED, handle_release)]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["--max-pdu", "123456"])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_ECHO
        assert events[1].event == evt.EVT_RELEASED
        requestor = events[1].assoc.requestor
        assert 123456 == requestor.maximum_length

    def test_flag_xe(self):
        """Test --request-little flag."""
        events = []

        def handle_echo(event):
            events.append(event)
            return 0x0000

        def handle_release(event):
            events.append(event)

        handlers = [(evt.EVT_C_ECHO, handle_echo), (evt.EVT_RELEASED, handle_release)]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-xe"])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_ECHO
        assert events[1].event == evt.EVT_RELEASED
        requestor = events[1].assoc.requestor
        cxs = requestor.primitive.presentation_context_definition_list
        assert len(cxs) == 1
        assert cxs[0].abstract_syntax == Verification
        assert cxs[0].transfer_syntax == [ExplicitVRLittleEndian]

    def test_flag_xb(self):
        """Test --request-big flag."""
        events = []

        def handle_echo(event):
            events.append(event)
            return 0x0000

        def handle_release(event):
            events.append(event)

        handlers = [(evt.EVT_C_ECHO, handle_echo), (evt.EVT_RELEASED, handle_release)]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-xb"])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_ECHO
        assert events[1].event == evt.EVT_RELEASED
        requestor = events[1].assoc.requestor
        cxs = requestor.primitive.presentation_context_definition_list
        assert len(cxs) == 1
        assert cxs[0].abstract_syntax == Verification
        assert cxs[0].transfer_syntax == [ExplicitVRBigEndian]

    def test_flag_xi(self):
        """Test --request-implicit flag."""
        events = []

        def handle_echo(event):
            events.append(event)
            return 0x0000

        def handle_release(event):
            events.append(event)

        handlers = [(evt.EVT_C_ECHO, handle_echo), (evt.EVT_RELEASED, handle_release)]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-xi"])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_ECHO
        assert events[1].event == evt.EVT_RELEASED
        requestor = events[1].assoc.requestor
        cxs = requestor.primitive.presentation_context_definition_list
        assert len(cxs) == 1
        assert cxs[0].abstract_syntax == Verification
        assert cxs[0].transfer_syntax == [ImplicitVRLittleEndian]

    def test_flag_repeat(self):
        """Test --repeat flag."""
        events = []

        def handle_echo(event):
            events.append(event)
            return 0x0000

        def handle_release(event):
            events.append(event)

        handlers = [(evt.EVT_C_ECHO, handle_echo), (evt.EVT_RELEASED, handle_release)]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["--repeat", "3"])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_ECHO
        assert events[1].event == evt.EVT_C_ECHO
        assert events[2].event == evt.EVT_C_ECHO
        assert events[3].event == evt.EVT_RELEASED

    def test_flag_abort(self):
        """Test --abort flag."""
        events = []

        def handle_echo(event):
            events.append(event)
            return 0x0000

        def handle_release(event):
            events.append(event)

        def handle_abort(event):
            events.append(event)

        handlers = [
            (evt.EVT_C_ECHO, handle_echo),
            (evt.EVT_RELEASED, handle_release),
            (evt.EVT_ABORTED, handle_abort),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["--abort"])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_ECHO
        assert events[1].event == evt.EVT_ABORTED


class TestEchoSCU(EchoSCUBase):
    """Tests for echoscu.py"""

    def setup(self):
        """Run prior to each test"""
        self.ae = None
        self.func = start_echoscu


class TestEchoSCUCLI(EchoSCUBase):
    """Tests for echoscu using CLI"""

    def setup(self):
        """Run prior to each test"""
        self.ae = None
        self.func = start_echoscu_cli



"""Hide modules from import

 ModuleHider - import finder hook and context manager
 hide_modules - decorator using ModuleHider

https://github.com/roryyorke/py-hide-modules
"""

# Copyright (c) 2019 Rory Yorke
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

try:
    import importlib.abc

    # py>=3.3 has MetaPathFinder
    _ModuleHiderBase = getattr(importlib.abc, "MetaPathFinder", importlib.abc.Finder)
except ImportError:
    # py2
    _ModuleHiderBase = object


class ModuleHider(_ModuleHiderBase):
    """Import finder hook to hide specified modules
    ModuleHider(hidden_modules) -> instance
    hidden_modules is a list of strings naming modules to hide.
    """

    def __init__(self, hidden):
        self.hidden = hidden
        self.hidden_modules = {}

    # python <=3.3
    def find_module(self, fullname, path=None):
        return self.find_spec(fullname, path)

    # python >=3.4
    def find_spec(self, fullname, path, target=None):
        if fullname in self.hidden:
            raise ImportError("No module named {}".format(fullname))

    def hide(self):
        "Starting hiding modules"
        import sys

        if self in sys.meta_path:
            raise RuntimeError("Already hiding modules")
        # must be first to override standard finders
        sys.meta_path.insert(0, self)
        # remove hidden modules to force reload
        for m in self.hidden:
            if m in sys.modules:
                self.hidden_modules[m] = sys.modules[m]
                del sys.modules[m]

    def unhide(self):
        "Unhide modules"
        import sys

        sys.meta_path.remove(self)
        sys.modules.update(self.hidden_modules)
        self.hidden_modules.clear()

    def __enter__(self):
        self.hide()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unhide()

    # there's much point in __del__: sys.meta_path will keep a
    # reference to an object on which .unhide() is not called, so
    # refcount will only go to 0 if the object is removed from
    # sys.meta_path somehow (in which case deletion doesn't
    # matter), or when Python exits (ditto)


def hide_modules(hidden):
    """hide_modules(hidden_modules) -> decorator

    When decorated function is called, the specified list of modules
    will be hidden; once the function exits, the modules will be
    unhidden.
    """

    def applydec(f):
        def decf(*args, **kwargs):
            with ModuleHider(hidden):
                f(*args, **kwargs)

        # carry across name so that nose still finds the test
        decf.__name__ = f.__name__
        # and carry across doc for test descriptions (etc.)
        decf.__doc__ = f.__doc__
        return decf

    return applydec



"""Test DIMSE service provider operations."""

from datetime import datetime
import queue
from io import BytesIO
import logging
import sys
import threading
import time

import pytest

from pydicom.dataset import Dataset

from pynetdicom import evt, AE, Association, _config, debug_logger
from pynetdicom.association import ServiceUser
from pynetdicom.dimse import DIMSEServiceProvider
from pynetdicom.dimse_messages import (
    C_STORE_RQ,
    C_STORE_RSP,
    C_FIND_RQ,
    C_FIND_RSP,
    C_GET_RQ,
    C_GET_RSP,
    C_MOVE_RQ,
    C_MOVE_RSP,
    C_ECHO_RQ,
    C_ECHO_RSP,
    C_CANCEL_RQ,
    N_EVENT_REPORT_RQ,
    N_EVENT_REPORT_RSP,
    N_GET_RQ,
    N_GET_RSP,
    N_SET_RQ,
    N_SET_RSP,
    N_ACTION_RQ,
    N_ACTION_RSP,
    N_CREATE_RQ,
    N_CREATE_RSP,
    N_DELETE_RQ,
    N_DELETE_RSP,
)
from pynetdicom.dimse_primitives import (
    C_STORE,
    C_ECHO,
    C_GET,
    C_MOVE,
    C_FIND,
    N_EVENT_REPORT,
    N_SET,
    N_GET,
    N_ACTION,
    N_CREATE,
    N_DELETE,
    C_CANCEL,
)
from pynetdicom.dsutils import encode
from pynetdicom.events import Event
from pynetdicom.pdu_primitives import P_DATA
from pynetdicom.pdu import P_DATA_TF
from .encoded_dimse_msg import c_store_ds
from .encoded_dimse_n_msg import (
    n_er_rq_ds,
    n_er_rsp_ds,
    n_get_rsp_ds,
    n_set_rq_ds,
    n_set_rsp_ds,
    n_action_rq_ds,
    n_action_rsp_ds,
    n_create_rq_ds,
    n_create_rsp_ds,
)
from .encoded_pdu_items import p_data_tf
from pynetdicom.sop_class import (
    Verification,
    BasicGrayscalePrintManagementMeta,
    Printer,
)


# debug_logger()


class DummyAssociation:
    def __init__(self):
        self.ae = AE()
        self.mode = None
        self.dul = DummyDUL()
        self.requestor = ServiceUser(self, "requestor")
        self.requestor.port = 11112
        self.requestor.ae_title = "TEST_LOCAL      "
        self.requestor.address = "127.0.0.1"
        self.requestor.maximum_length = 31682
        self.acceptor = ServiceUser(self, "acceptor")
        self.acceptor.ae_title = "TEST_REMOTE     "
        self.acceptor.port = 11113
        self.acceptor.address = "127.0.0.2"
        self.acse_timeout = 11
        self.dimse_timeout = 1
        self.network_timeout = 13
        self.is_killed = False
        self.is_aborted = False
        self.is_established = False
        self.is_rejected = False
        self.is_released = False
        self.is_acceptor = False
        self.is_requestor = True
        self._handlers = {}

    def abort(self):
        self.is_aborted = True
        self.kill()

    def kill(self):
        self.is_killed = True

    @property
    def requested_contexts(self):
        return self.requestor.get_contexts("requested")

    @property
    def supported_contexts(self):
        return self.requestor.get_contexts("supported")

    def get_handlers(self, event):
        if event not in self._handlers:
            return []

        return self._handlers[event]


class DummyDUL:
    """Dummy DUL class for testing DIMSE provider"""

    def __init__(self):
        self.event_queue = queue.Queue()

    @staticmethod
    def is_alive():
        return True

    @staticmethod
    def send_pdu(pdv):
        """Dummy Send method to test DIMSEServiceProvider.Send"""
        pass

    @staticmethod
    def receive_pdu():
        """Dummy Receive method to test DIMSEServiceProvider.Receive"""
        pass

    @staticmethod
    def peek_next_pdu():
        return 0x01


REFERENCE_MSG = [
    (C_ECHO(), ("C_ECHO_RQ", "C_ECHO_RSP")),
    (C_STORE(), ("C_STORE_RQ", "C_STORE_RSP")),
    (C_FIND(), ("C_FIND_RQ", "C_FIND_RSP")),
    (C_GET(), ("C_GET_RQ", "C_GET_RSP")),
    (C_MOVE(), ("C_MOVE_RQ", "C_MOVE_RSP")),
    (C_CANCEL(), (None, "C_CANCEL_RQ")),
    (N_EVENT_REPORT(), ("N_EVENT_REPORT_RQ", "N_EVENT_REPORT_RSP")),
    (N_GET(), ("N_GET_RQ", "N_GET_RSP")),
    (N_SET(), ("N_SET_RQ", "N_SET_RSP")),
    (N_ACTION(), ("N_ACTION_RQ", "N_ACTION_RSP")),
    (N_CREATE(), ("N_CREATE_RQ", "N_CREATE_RSP")),
    (N_DELETE(), ("N_DELETE_RQ", "N_DELETE_RSP")),
]


class TestDIMSEProvider:
    """Test DIMSE service provider operations."""

    def setup(self):
        """Set up"""
        self.dimse = DIMSEServiceProvider(DummyAssociation())

    def test_receive_not_pdata(self):
        """Test we get back None if not a P_DATA"""
        assert self.dimse.get_msg(True) == (None, None)

    def test_peek_empty(self):
        """Test peek_msg with nothing on the queue."""
        dimse = DIMSEServiceProvider(DummyAssociation())
        assert dimse.peek_msg() == (None, None)

    def test_peek_item(self):
        """Test peek_msg with nothing on the queue."""
        dimse = DIMSEServiceProvider(DummyAssociation())
        primitive = C_STORE()
        dimse.msg_queue.put((14, primitive))
        assert dimse.peek_msg() == (14, primitive)

    def test_invalid_message(self, monkeypatch):
        """Test that an invalid message kills the association."""

        monkeypatch.setattr(_config, "ENFORCE_UID_CONFORMANCE", True)

        dimse = DIMSEServiceProvider(DummyAssociation())

        p_data_tf = (
            #     |   | length
            b"\x04\x00\x00\x00\x00\x4E"  # P-DATA-TF 78
            b"\x00\x00\x00\x4a\x01"  # PDV Item 70
            b"\x03"  # PDV: 2 -> 69
            # C-ECHO-RQ
            # CommandGroupLen | len 4         | value 64
            b"\x00\x00\x00\x00\x04\x00\x00\x00\x40\x00\x00\x00"  # 12
            #  AffSOPClass    | len 18        | 1.2.840.00008.1.1 - invalid
            b"\x00\x00\x02\x00\x12\x00\x00\x00"
            b"\x31\x2e\x32\x2e\x38\x34\x30\x2e\x30\x30\x30\x30\x38\x2e\x31\x2e"
            b"\x31\x00"  # 26
            # CommandField    | len 2         | 0x0030 -> C-ECHO-RQ
            b"\x00\x00\x00\x01\x02\x00\x00\x00\x30\x00"  # 10
            # Message ID      | len 6         | value
            b"\x00\x00\x10\x01\x06\x00\x00\x00\xff\xff\xff\xff\xff\xff"  # 14
            # CommandDSType   | len 2         | no DS
            b"\x00\x00\x00\x08\x02\x00\x00\x00\x01\x01"  # 10
        )
        pdata = P_DATA_TF()
        pdata.decode(p_data_tf)
        pdata = pdata.to_primitive()
        # Should send Evt19 due to invalid message
        dimse.receive_primitive(pdata)
        assert dimse.assoc.dul.event_queue.get() == "Evt19"


class TestEventHandlingAcceptor:
    """Test the transport events and handling as acceptor."""

    def setup(self):
        self.ae = None
        _config.LOG_HANDLER_LEVEL = "none"

    def teardown(self):
        if self.ae:
            self.ae.shutdown()

        _config.LOG_HANDLER_LEVEL = "standard"

    def test_no_handlers(self):
        """Test with no transport event handlers bound."""
        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == []
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == []
        assoc = ae.associate("localhost", 11112)

        assert assoc.is_established
        assert len(scp.active_associations) == 1
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == []
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == []
        assert assoc.get_handlers(evt.EVT_DIMSE_RECV) == []
        assert assoc.get_handlers(evt.EVT_DIMSE_SENT) == []

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_RECV) == []
        assert child.get_handlers(evt.EVT_DIMSE_SENT) == []

        assoc.release()
        scp.shutdown()

    def test_dimse_sent(self):
        """Test binding to EVT_DIMSE_SENT."""
        triggered = []

        def handle(event):
            triggered.append(event)

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_SENT, handle)]
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == [(handle, None)]

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        assert len(scp.active_associations) == 1
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == [(handle, None)]
        assert assoc.get_handlers(evt.EVT_DIMSE_SENT) == []

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_SENT) == [(handle, None)]

        assoc.send_c_echo()

        assoc.release()

        while scp.active_associations:
            time.sleep(0.05)

        assert len(triggered) == 1
        event = triggered[0]
        assert isinstance(event, Event)
        assert isinstance(event.assoc, Association)
        assert isinstance(event.timestamp, datetime)
        assert event.event.name == "EVT_DIMSE_SENT"

        assert isinstance(triggered[0].message, C_ECHO_RSP)

        scp.shutdown()

    def test_dimse_sent_bind(self):
        """Test binding to EVT_DIMSE_SENT."""
        triggered = []

        def handle(event):
            triggered.append(event)

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_SENT, handle)]
        scp = ae.start_server(("localhost", 11112), block=False)
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == []

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assoc.send_c_echo(msg_id=12)

        scp.bind(evt.EVT_DIMSE_SENT, handle)

        assert len(scp.active_associations) == 1
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == [(handle, None)]
        assert assoc.get_handlers(evt.EVT_DIMSE_SENT) == []

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_SENT) == [(handle, None)]

        assoc.send_c_echo(msg_id=21)

        assoc.release()

        while scp.active_associations:
            time.sleep(0.05)

        assert len(triggered) == 1
        event = triggered[0]
        assert isinstance(event, Event)
        assert isinstance(event.assoc, Association)
        assert isinstance(event.timestamp, datetime)
        assert event.event.name == "EVT_DIMSE_SENT"

        assert isinstance(triggered[0].message, C_ECHO_RSP)
        assert triggered[0].message.command_set.MessageIDBeingRespondedTo == 21

        scp.shutdown()

    def test_dimse_sent_unbind(self):
        """Test unbinding EVT_DIMSE_SENT."""
        triggered = []

        def handle(event):
            triggered.append(event)

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_SENT, handle)]
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == [(handle, None)]

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_SENT) == [(handle, None)]

        assert assoc.get_handlers(evt.EVT_DIMSE_SENT) == []

        assoc.send_c_echo(msg_id=12)

        scp.unbind(evt.EVT_DIMSE_SENT, handle)

        assert len(scp.active_associations) == 1
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == []

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_SENT) == []

        assoc.send_c_echo(msg_id=21)

        assoc.release()

        while scp.active_associations:
            time.sleep(0.05)

        assert len(triggered) == 1
        event = triggered[0]
        assert isinstance(event, Event)
        assert isinstance(event.assoc, Association)
        assert isinstance(event.timestamp, datetime)
        assert event.event.name == "EVT_DIMSE_SENT"

        assert isinstance(triggered[0].message, C_ECHO_RSP)
        assert triggered[0].message.command_set.MessageIDBeingRespondedTo == 12

        scp.shutdown()

    def test_dimse_sent_raises(self, caplog):
        """Test the handler for EVT_DIMSE_SENT raising exception."""

        def handle(event):
            raise NotImplementedError("Exception description")

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_SENT, handle)]
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        with caplog.at_level(logging.ERROR, logger="pynetdicom"):
            assoc = ae.associate("localhost", 11112)
            assert assoc.is_established
            assoc.send_c_echo()
            assoc.release()

            while scp.active_associations:
                time.sleep(0.05)

            scp.shutdown()

            msg = (
                "Exception raised in user's 'evt.EVT_DIMSE_SENT' event handler"
                " 'handle'"
            )
            assert msg in caplog.text
            assert "Exception description" in caplog.text

    def test_dimse_recv(self):
        """Test starting bound to EVT_DIMSE_RECV."""
        triggered = []

        def handle(event):
            triggered.append(event)

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_RECV, handle)]
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == [(handle, None)]

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        assert len(scp.active_associations) == 1
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == [(handle, None)]
        assert assoc.get_handlers(evt.EVT_DIMSE_RECV) == []

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_RECV) == [(handle, None)]

        assoc.send_c_echo()

        assoc.release()

        while scp.active_associations:
            time.sleep(0.05)

        assert len(triggered) == 1
        event = triggered[0]
        assert isinstance(event, Event)
        assert isinstance(event.assoc, Association)
        assert isinstance(event.timestamp, datetime)
        assert isinstance(triggered[0].message, C_ECHO_RQ)
        assert event.event.name == "EVT_DIMSE_RECV"

        scp.shutdown()

    def test_dimse_recv_bind(self):
        """Test binding to EVT_DIMSE_RECV."""
        triggered = []

        def handle(event):
            triggered.append(event)

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == []

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        assert len(scp.active_associations) == 1

        assoc.send_c_echo(msg_id=12)

        scp.bind(evt.EVT_DIMSE_RECV, handle)

        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == [(handle, None)]
        assert assoc.get_handlers(evt.EVT_DIMSE_RECV) == []

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_RECV) == [(handle, None)]

        assoc.send_c_echo(msg_id=21)

        assoc.release()

        while scp.active_associations:
            time.sleep(0.05)

        assert len(triggered) == 1
        event = triggered[0]
        assert isinstance(event, Event)
        assert isinstance(event.assoc, Association)
        assert isinstance(event.timestamp, datetime)
        assert isinstance(triggered[0].message, C_ECHO_RQ)
        assert event.event.name == "EVT_DIMSE_RECV"
        assert triggered[0].message.command_set.MessageID == 21

        scp.shutdown()

    def test_dimse_recv_unbind(self):
        """Test unbinding to EVT_DIMSE_RECV."""
        triggered = []

        def handle(event):
            triggered.append(event)

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_RECV, handle)]
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == [(handle, None)]

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        assert len(scp.active_associations) == 1

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_RECV) == [(handle, None)]

        assoc.send_c_echo(msg_id=12)

        scp.unbind(evt.EVT_DIMSE_RECV, handle)

        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == []
        assert assoc.get_handlers(evt.EVT_DIMSE_RECV) == []

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_RECV) == []

        assoc.send_c_echo(msg_id=21)

        assoc.release()

        while scp.active_associations:
            time.sleep(0.05)

        assert len(triggered) == 1
        event = triggered[0]
        assert isinstance(event, Event)
        assert isinstance(event.assoc, Association)
        assert isinstance(event.timestamp, datetime)
        assert isinstance(triggered[0].message, C_ECHO_RQ)
        assert event.event.name == "EVT_DIMSE_RECV"
        assert triggered[0].message.command_set.MessageID == 12

        scp.shutdown()

    def test_dimse_recv_raises(self, caplog):
        """Test the handler for EVT_DIMSE_RECV raising exception."""

        def handle(event):
            raise NotImplementedError("Exception description")

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_RECV, handle)]
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        with caplog.at_level(logging.ERROR, logger="pynetdicom"):
            assoc = ae.associate("localhost", 11112)
            assert assoc.is_established
            assoc.send_c_echo()
            assoc.release()

            while scp.active_associations:
                time.sleep(0.05)

            scp.shutdown()

            msg = (
                "Exception raised in user's 'evt.EVT_DIMSE_RECV' event handler"
                " 'handle'"
            )
            assert msg in caplog.text
            assert "Exception description" in caplog.text


class TestEventHandlingRequestor:
    """Test the transport events and handling as requestor."""

    def setup(self):
        self.ae = None
        _config.LOG_HANDLER_LEVEL = "none"

    def teardown(self):
        if self.ae:
            self.ae.shutdown()

        _config.LOG_HANDLER_LEVEL = "standard"

    def test_no_handlers(self):
        """Test with no transport event handlers bound."""
        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == []
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == []
        assoc = ae.associate("localhost", 11112)

        assert assoc.is_established
        assert len(scp.active_associations) == 1
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == []
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == []
        assert assoc.get_handlers(evt.EVT_DIMSE_RECV) == []
        assert assoc.get_handlers(evt.EVT_DIMSE_SENT) == []

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_RECV) == []
        assert child.get_handlers(evt.EVT_DIMSE_SENT) == []

        assoc.release()
        scp.shutdown()

    def test_dimse_sent(self):
        """Test binding to EVT_DIMSE_SENT."""
        triggered = []

        def handle(event):
            triggered.append(event)

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_SENT, handle)]
        scp = ae.start_server(("localhost", 11112), block=False)
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == []

        assoc = ae.associate("localhost", 11112, evt_handlers=handlers)
        assert assoc.is_established
        assert len(scp.active_associations) == 1
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == []
        assert assoc.get_handlers(evt.EVT_DIMSE_SENT) == [(handle, None)]

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_SENT) == []

        assoc.send_c_echo()
        assoc.release()

        while scp.active_associations:
            time.sleep(0.05)

        assert len(triggered) == 1
        event = triggered[0]
        assert isinstance(event, Event)
        assert isinstance(event.assoc, Association)
        assert isinstance(event.timestamp, datetime)
        assert event.event.name == "EVT_DIMSE_SENT"

        assert isinstance(triggered[0].message, C_ECHO_RQ)

        scp.shutdown()

    def test_dimse_sent_bind(self):
        """Test binding to EVT_DIMSE_SENT."""
        triggered = []

        def handle(event):
            triggered.append(event)

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_SENT, handle)]
        scp = ae.start_server(("localhost", 11112), block=False)
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == []

        assoc = ae.associate("localhost", 11112)
        assert assoc.get_handlers(evt.EVT_DIMSE_SENT) == []
        assert assoc.is_established
        assoc.send_c_echo(msg_id=12)

        assoc.bind(evt.EVT_DIMSE_SENT, handle)

        assert len(scp.active_associations) == 1
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == []
        assert assoc.get_handlers(evt.EVT_DIMSE_SENT) == [(handle, None)]

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_SENT) == []

        assoc.send_c_echo(msg_id=21)

        assoc.release()

        while scp.active_associations:
            time.sleep(0.05)

        assert len(triggered) == 1
        event = triggered[0]
        assert isinstance(event, Event)
        assert isinstance(event.assoc, Association)
        assert isinstance(event.timestamp, datetime)
        assert event.event.name == "EVT_DIMSE_SENT"

        assert isinstance(triggered[0].message, C_ECHO_RQ)
        assert triggered[0].message.command_set.MessageID == 21

        scp.shutdown()

    def test_dimse_sent_unbind(self):
        """Test unbinding EVT_DIMSE_SENT."""
        triggered = []

        def handle(event):
            triggered.append(event)

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_SENT, handle)]
        scp = ae.start_server(("localhost", 11112), block=False)
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == []

        assoc = ae.associate("localhost", 11112, evt_handlers=handlers)
        assert assoc.is_established
        assoc.send_c_echo(msg_id=12)
        assert len(scp.active_associations) == 1
        assert scp.get_handlers(evt.EVT_DIMSE_SENT) == []
        assert assoc.get_handlers(evt.EVT_DIMSE_SENT) == [(handle, None)]

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_SENT) == []

        assoc.unbind(evt.EVT_DIMSE_SENT, handle)

        assert assoc.get_handlers(evt.EVT_DIMSE_SENT) == []

        assoc.send_c_echo(msg_id=21)

        assoc.release()

        while scp.active_associations:
            time.sleep(0.05)

        assert len(triggered) == 1
        assert triggered[0].message.command_set.MessageID == 12

        scp.shutdown()

    def test_dimse_sent_raises(self, caplog):
        """Test the handler for EVT_DIMSE_SENT raising exception."""

        def handle(event):
            raise NotImplementedError("Exception description")

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_SENT, handle)]
        scp = ae.start_server(("localhost", 11112), block=False)

        with caplog.at_level(logging.ERROR, logger="pynetdicom"):
            assoc = ae.associate("localhost", 11112, evt_handlers=handlers)
            assert assoc.is_established
            assoc.send_c_echo()
            assoc.release()

            while scp.active_associations:
                time.sleep(0.05)

            scp.shutdown()

            msg = (
                "Exception raised in user's 'evt.EVT_DIMSE_SENT' event handler"
                " 'handle'"
            )
            assert msg in caplog.text
            assert "Exception description" in caplog.text

    def test_dimse_recv(self):
        """Test starting bound to EVT_DIMSE_RECV."""
        triggered = []

        def handle(event):
            triggered.append(event)

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_RECV, handle)]
        scp = ae.start_server(("localhost", 11112), block=False)
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == []

        assoc = ae.associate("localhost", 11112, evt_handlers=handlers)
        assert assoc.is_established
        assert len(scp.active_associations) == 1
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == []
        assert assoc.get_handlers(evt.EVT_DIMSE_RECV) == [(handle, None)]

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_RECV) == []

        assoc.send_c_echo()

        assoc.release()

        while scp.active_associations:
            time.sleep(0.05)

        assert len(triggered) == 1
        event = triggered[0]
        assert isinstance(event, Event)
        assert isinstance(event.assoc, Association)
        assert isinstance(event.timestamp, datetime)
        assert isinstance(triggered[0].message, C_ECHO_RSP)
        assert event.event.name == "EVT_DIMSE_RECV"

        scp.shutdown()

    def test_dimse_recv_bind(self):
        """Test binding to EVT_DIMSE_RECV."""
        triggered = []

        def handle(event):
            triggered.append(event)

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == []

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        assert len(scp.active_associations) == 1
        assert assoc.get_handlers(evt.EVT_DIMSE_RECV) == []
        assoc.send_c_echo(msg_id=12)

        assoc.bind(evt.EVT_DIMSE_RECV, handle)

        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == []
        assert assoc.get_handlers(evt.EVT_DIMSE_RECV) == [(handle, None)]

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_RECV) == []

        assoc.send_c_echo(msg_id=21)

        assoc.release()

        while scp.active_associations:
            time.sleep(0.05)

        assert len(triggered) == 1
        event = triggered[0]
        assert isinstance(event, Event)
        assert isinstance(event.assoc, Association)
        assert isinstance(event.timestamp, datetime)
        assert isinstance(triggered[0].message, C_ECHO_RSP)
        assert event.event.name == "EVT_DIMSE_RECV"
        assert triggered[0].message.command_set.MessageIDBeingRespondedTo == 21

        scp.shutdown()

    def test_dimse_recv_unbind(self):
        """Test unbinding to EVT_DIMSE_RECV."""
        triggered = []

        def handle(event):
            triggered.append(event)

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_RECV, handle)]
        scp = ae.start_server(("localhost", 11112), block=False)
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == []

        assoc = ae.associate("localhost", 11112, evt_handlers=handlers)
        assert assoc.is_established
        assert assoc.get_handlers(evt.EVT_DIMSE_RECV) == [(handle, None)]
        assoc.send_c_echo(msg_id=12)

        assoc.unbind(evt.EVT_DIMSE_RECV, handle)

        assert len(scp.active_associations) == 1
        assert scp.get_handlers(evt.EVT_DIMSE_RECV) == []
        assert assoc.get_handlers(evt.EVT_DIMSE_RECV) == []

        child = scp.active_associations[0]
        assert child.get_handlers(evt.EVT_DIMSE_RECV) == []

        assoc.send_c_echo(msg_id=21)

        assoc.release()

        while scp.active_associations:
            time.sleep(0.05)

        assert len(triggered) == 1
        event = triggered[0]
        assert isinstance(event, Event)
        assert isinstance(event.assoc, Association)
        assert isinstance(event.timestamp, datetime)
        assert isinstance(triggered[0].message, C_ECHO_RSP)
        assert event.event.name == "EVT_DIMSE_RECV"
        assert triggered[0].message.command_set.MessageIDBeingRespondedTo == 12

        scp.shutdown()

    def test_dimse_recv_raises(self, caplog):
        """Test the handler for EVT_DIMSE_RECV raising exception."""

        def handle(event):
            raise NotImplementedError("Exception description")

        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        ae.add_requested_context(Verification)
        handlers = [(evt.EVT_DIMSE_RECV, handle)]
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        with caplog.at_level(logging.ERROR, logger="pynetdicom"):
            assoc = ae.associate("localhost", 11112)
            assert assoc.is_established
            assoc.send_c_echo()
            assoc.release()

            while scp.active_associations:
                time.sleep(0.05)

            scp.shutdown()

            msg = (
                "Exception raised in user's 'evt.EVT_DIMSE_RECV' event handler"
                " 'handle'"
            )
            assert msg in caplog.text
            assert "Exception description" in caplog.text



"""Tests for the _validators module"""

import pytest


from pynetdicom import _config
from pynetdicom._validators import validate_ae, validate_ui


AE_REFERENCE = [
    ("", True, ""),
    ("A", True, ""),
    ("Z", True, ""),
    ("a", True, ""),
    ("z", True, ""),
    ("1", True, ""),
    ("0", True, ""),
    ("aaaaaaaaaaaaaaaa", True, ""),
    ("               a", True, ""),
    ("a               ", True, ""),
    ("        a       ", True, ""),
    ("                ", True, ""),
    ("                 ", False, "must not exceed 16 characters"),
    ("aaaaaaaaaaaaaaaaa", False, "must not exceed 16 characters"),
    ("\\", False, "must not contain control characters or backslashes"),
    ("\n", False, "must not contain control characters or backslashes"),
    ("\t", False, "must not contain control characters or backslashes"),
    ("\r", False, "must not contain control characters or backslashes"),
    (b"A", False, "must be str"),
    # zero-width space
    ("\u200b5", False, "must only contain ASCII characters"),
]


@pytest.mark.parametrize("value, ref_result, ref_msg", AE_REFERENCE)
def test_validate_ae(value, ref_result, ref_msg):
    """Tests for validate_ae()"""
    result, msg = validate_ae(value)
    assert result == ref_result
    assert ref_msg == msg


UI_REFERENCE = [
    # value, result if enforcing conf, result if not enforcing conf
    ("", (False, "UID is non-conformant"), (False, "must not be an empty str")),
    (" ", (False, "UID is non-conformant"), (False, "must not be an empty str")),
    ("a", (False, "UID is non-conformant"), (True, "")),
    (
        "a" * 65,
        (False, "UID is non-conformant"),
        (False, "must not exceed 64 characters"),
    ),
    ("1", (True, ""), (True, "")),
    (" 1", (True, ""), (True, "")),
    (b"1", (False, "must be pydicom.uid.UID"), (False, "must be pydicom.uid.UID")),
    ("1.2.03", (False, "UID is non-conformant"), (True, "")),
    ("1.2.840.10008.1.2", (True, ""), (True, "")),
]


@pytest.fixture()
def enforce_uid_conformance():
    _config.ENFORCE_UID_CONFORMANCE = True
    yield
    _config.ENFORCE_UID_CONFORMANCE = False


@pytest.mark.parametrize("value, conf, nonconf", UI_REFERENCE)
def test_validate_ui_conf(value, conf, nonconf, enforce_uid_conformance):
    """Tests for validate_ui() if enforcing conformance"""
    assert validate_ui(value) == conf


@pytest.mark.parametrize("value, conf, nonconf", UI_REFERENCE)
def test_validate_ui_nonconf(value, conf, nonconf):
    """Tests for validate_ui() if not enforcing conformance"""
    assert validate_ui(value) == nonconf



import time


def sleep(duration):
    """Sleep for at least `duration` seconds."""
    now = time.perf_counter()
    end = now + duration
    while now < end:
        now = time.perf_counter()


def wait_for_server_socket(server, timeout=5):
    """Sleep until the AssociationServer's socket is bound to an address.

    Parameters
    ----------
    server : pynetdicom.transport.AssociationServer
        The server that we're waiting on.
    timeout : int or float, optional
        The maximum number of seconds to wait for.
    """
    # TODO: Python >= 3.8: use sys.audit() instead
    timeout = 0
    while timeout < 1:
        try:
            assert server.socket.getsockname() != ("0.0.0.0", 0)
            return
        except AssertionError:
            time.sleep(0.05)
            timeout += 0.05



"""Status testing"""

import logging

import pytest

from pydicom.dataset import Dataset

from pynetdicom.status import code_to_category, code_to_status

try:
    from pynetdicom.status import Status

    HAS_STATUS = True
except ImportError:
    HAS_STATUS = False


LOGGER = logging.getLogger("pynetdicom")
LOGGER.setLevel(logging.CRITICAL)


class TestStatus:
    """Test the status.py module"""

    def test_code_to_status(self):
        """Test converting a status code to a Dataset"""
        status = code_to_status(0x0123)
        assert isinstance(status, Dataset)
        assert status.Status == 0x0123
        del status.Status
        assert status == Dataset()

        assert pytest.raises(ValueError, code_to_status, -0x0001)

    def test_code_to_category_unknown(self):
        """Test converting an unknown status code to its category"""
        assert code_to_category(0xDF01) == "Unknown"

    def test_code_to_category_general(self):
        """Test converting a general status code to its category"""
        c2c = code_to_category

        # General status
        assert c2c(0x0000) == "Success"
        assert c2c(0xFF00) == "Pending"
        assert c2c(0xFF01) == "Pending"
        assert c2c(0xFE00) == "Cancel"

        assert c2c(0x0107) == "Warning"

    def test_code_to_category_negative_raises(self):
        """Test code_to_category raises exception if negative."""
        assert pytest.raises(ValueError, code_to_category, -0x0001)

    def test_code_to_category_storage(self):
        """Test converting storage service class status code to its category"""
        # Storage Service Class specific PS3.4 Annex B
        c2c = code_to_category

        for code in range(0xA700, 0xA800):
            assert c2c(code) == "Failure"
        for code in range(0xA900, 0xAA00):
            assert c2c(code) == "Failure"
        for code in range(0xC000, 0xD000):
            assert c2c(code) == "Failure"
        for code in [0xB000, 0xB007, 0xB006]:
            assert c2c(code) == "Warning"
        assert c2c(0x0000) == "Success"

    def test_code_to_category_qr_find(self):
        """Test converting Q/R find service class status code to its category"""
        # Query/Retrieve Service Class specific - Find, PS3.4 Annex C.4.1
        c2c = code_to_category

        assert c2c(0xA700) == "Failure"
        assert c2c(0xA900) == "Failure"
        for code in range(0xC000, 0xD000):
            assert c2c(code) == "Failure"
        assert c2c(0xFE00) == "Cancel"
        for code in [0xFF00, 0xFF01]:
            assert c2c(code) == "Pending"
        assert c2c(0x0000) == "Success"

    def test_code_to_category_qr_move(self):
        """Test converting Q/R move service class status code to its category"""
        # Query/Retrieve Service Class specific - Move, PS3.4 Annex C.4.2
        c2c = code_to_category

        for code in [0xA701, 0xA702, 0xA801, 0xA900]:
            assert c2c(code) == "Failure"
        for code in range(0xC000, 0xD000):
            assert c2c(code) == "Failure"
        assert c2c(0xFE00) == "Cancel"
        assert c2c(0xB000) == "Warning"
        for code in [0xFF00]:
            assert c2c(code) == "Pending"
        assert c2c(0x0000) == "Success"

    def test_code_to_category_qr_get(self):
        """Test converting Q/R get service class status code to its category"""
        # Query/Retrieve Service Class specific - Get, PS3.4 Annex C.4.3
        c2c = code_to_category

        for code in [0xA701, 0xA702, 0xA900]:
            assert c2c(code) == "Failure"
        for code in range(0xC000, 0xD000):
            assert c2c(code) == "Failure"
        assert c2c(0xFE00) == "Cancel"
        assert c2c(0xB000) == "Warning"
        for code in [0xFF00]:
            assert c2c(code) == "Pending"
        assert c2c(0x0000) == "Success"

    def test_code_to_category_proc_step(self):
        """Test converting procedure step status code to its category"""
        # Procedure Step, PS3.4 Annex F
        c2c = code_to_category

        assert c2c(0x0110) == "Failure"
        assert c2c(0x0001) == "Warning"

    def test_code_to_category_print_management(self):
        """Test converting print management status code to its category"""
        # Procedure Step, PS3.4 Annex H
        c2c = code_to_category

        assert c2c(0x0000) == "Success"
        for code in [0xB600, 0xB601, 0xB602, 0xB603, 0xB604, 0xB605, 0xB609, 0xB60A]:
            assert c2c(code) == "Warning"
        for code in [0xC600, 0xC601, 0xC602, 0xC603, 0xC605, 0xC613, 0xC616]:
            assert c2c(code) == "Failure"

    def test_code_to_category_basic_worklist(self):
        """Test converting basic worklist status code to its category"""
        # Basic Worklist Management, PS3.4 Annex K
        c2c = code_to_category

        assert c2c(0x0000) == "Success"
        for code in [0xA700, 0xA900]:
            assert c2c(code) == "Failure"
        for code in range(0xC000, 0xD000):
            assert c2c(code) == "Failure"
        assert c2c(0xFE00) == "Cancel"
        assert c2c(0xB000) == "Warning"
        for code in [0xFF00, 0xFF01]:
            assert c2c(code) == "Pending"

    def test_code_to_category_application_event(self):
        """Test converting application event status code to its category"""
        # Application Event Logging, PS3.4 Annex P
        c2c = code_to_category

        assert c2c(0x0000) == "Success"
        for code in [0xC101, 0xC102, 0xC103, 0xC104, 0xC10E, 0xC110, 0xC111]:
            assert c2c(code) == "Failure"
        for code in [0xB101, 0xB102, 0xB104]:
            assert c2c(code) == "Warning"

    def test_code_to_category_relevant_patient(self):
        """Test converting relevant patient status code to its category"""
        # Relevant Patient Information Query, PS3.4 Annex Q
        c2c = code_to_category

        assert c2c(0x0000) == "Success"
        for code in [0xA700, 0xA900, 0xC000, 0xC100, 0xC200]:
            assert c2c(code) == "Failure"
        assert c2c(0xFE00) == "Cancel"
        assert c2c(0xB000) == "Warning"
        for code in [0xFF00]:
            assert c2c(code) == "Pending"

    def test_code_to_category_media_creation(self):
        """Test converting media creation status code to its category"""
        # Media Creation Management, PS3.4 Annex S
        c2c = code_to_category

        for code in [0xA510, 0xC201, 0xC202, 0xC203]:
            assert c2c(code) == "Failure"
        assert c2c(0x0001) == "Warning"

    def test_code_to_category_substance_admin(self):
        """Test converting substance admin status code to its category"""
        # Substance Administration Query, PS3.4 Annex V
        c2c = code_to_category

        assert c2c(0x0000) == "Success"
        for code in [0xA700, 0xA900]:
            assert c2c(code) == "Failure"
        for code in range(0xC000, 0xD000):
            assert c2c(code) == "Failure"
        assert c2c(0xFE00) == "Cancel"
        assert c2c(0xB000) == "Warning"
        for code in [0xFF00, 0xFF01]:
            assert c2c(code) == "Pending"

    def test_code_to_category_instance_frame(self):
        """Test converting instance and frame status code to its category"""
        # Instance and Frame Level Retrieve, PS3.4 Annex Y
        c2c = code_to_category

        assert c2c(0x0000) == "Success"
        for code in [
            0xA701,
            0xA702,
            0xA801,
            0xA900,
            0xAA00,
            0xAA01,
            0xAA02,
            0xAA03,
            0xAA04,
        ]:
            assert c2c(code) == "Failure"
        for code in range(0xC000, 0xD000):
            assert c2c(code) == "Failure"
        assert c2c(0xFE00) == "Cancel"
        assert c2c(0xB000) == "Warning"
        for code in [0xFF00]:
            assert c2c(code) == "Pending"

    def test_code_to_category_composite_instance(self):
        """Test converting composite instance status code to its category"""
        # Composite Instance Retrieve, PS3.4 Annex Z
        c2c = code_to_category

        assert c2c(0x0000) == "Success"
        for code in [0xA701, 0xA702, 0xA900]:
            assert c2c(code) == "Failure"
        for code in range(0xC000, 0xD000):
            assert c2c(code) == "Failure"
        assert c2c(0xFE00) == "Cancel"
        assert c2c(0xB000) == "Warning"
        for code in [0xFF00]:
            assert c2c(code) == "Pending"

    def test_code_to_category_unified_procedure(self):
        """Test converting unified procedure status code to its category"""
        # Unified Procedure Step, PS3.4 Annex CC
        c2c = code_to_category

        assert c2c(0x0000) == "Success"
        for code in [0x0001, 0xB300, 0xB301, 0xB304, 0xB305, 0xB306]:
            assert c2c(code) == "Warning"
        for code in [
            0xC300,
            0xC301,
            0xC302,
            0xC304,
            0xC307,
            0xC310,
            0xC311,
            0xC313,
            0xC312,
            0xC308,
            0xC314,
            0xC315,
            0xC309,
            0xA700,
            0xA900,
            0x0122,
        ]:
            assert c2c(code) == "Failure"
        for code in range(0xC000, 0xD000):
            assert c2c(code) == "Failure"
        assert c2c(0xFE00) == "Cancel"
        for code in [0xFF00, 0xFF01]:
            assert c2c(code) == "Pending"

    def test_code_to_category_rt_machine(self):
        """Test converting RT machine status code to its category"""
        # RT Machine Verification, PS3.4 Annex DD
        c2c = code_to_category

        assert c2c(0x0000) == "Success"
        for code in [0xC227, 0xC221, 0xC222, 0xC223, 0xC224, 0xC225, 0xC226, 0xC112]:
            assert c2c(code) == "Failure"

    def test_code_to_category_non_patient(self):
        """Test converting non-patient object status code to its category"""
        # Non-patient Object Storage, PS3.4 Annex GG
        c2c = code_to_category

        assert c2c(0x0000) == "Success"
        for code in [0xA700, 0xA900, 0xC000]:
            assert c2c(code) == "Failure"


@pytest.mark.skipif(not HAS_STATUS, reason="No Status class available")
class TestStatusEnum:
    """Tests for the Status enum class."""

    def test_default(self):
        """Test the default class."""
        assert 0x0000 == Status.SUCCESS
        assert 0xFE00 == Status.CANCEL
        assert 0xFF00 == Status.PENDING

    def test_adding(self):
        """Tests for adding a new constant to the Status enum."""
        with pytest.raises(AttributeError, match=r"PENDING_WITH_WARNING"):
            Status.PENDING_WITH_WARNING

        Status.add("PENDING_WITH_WARNING", 0xFF01)
        assert 0xFF01 == Status.PENDING_WITH_WARNING
        assert 0x0000 == Status.SUCCESS
        assert 0xFE00 == Status.CANCEL
        assert 0xFF00 == Status.PENDING
